@@ -5,15 +5,18 @@
 
 struct game_s;
 
-#define AUDIENCE_STR_MAX    7
-#define AUDIENCE_BC_MAX     5
-#define AUDIENCE_BUF2_POS   320
-#define AUDIENCE_CBUF_POS   640
+#define AUDIENCE_STR_MAX 7
+#define AUDIENCE_BC_MAX 5
+#define AUDIENCE_DIPLO_MSG_SIZE 320
+#define AUDIENCE_STRTBL_BUFSIZE 640
+#define AUDIENCE_BUF2_POS 320
+#define AUDIENCE_CBUF_POS 640
 
 struct audience_s {
     struct game_s *g;
     void *uictx;
-    char *buf;
+    const char *buf;
+    char diplo_msg[AUDIENCE_DIPLO_MSG_SIZE];
     player_id_t ph;
     player_id_t pa;
     player_id_t pwar;
@@ -28,8 +31,9 @@ struct audience_s {
     int16_t tribute_bc;
     tech_field_t tribute_field;
     uint8_t tribute_tech;
+    char strtbl_buf[AUDIENCE_STRTBL_BUFSIZE];
     const char *strtbl[AUDIENCE_STR_MAX];
-    bool *condtbl;
+    const bool *condtbl;
     uint16_t bctbl[AUDIENCE_BC_MAX];
 };
 
