@@ -154,6 +154,7 @@ static const struct numtbl_s {
     DEFNUMITEMLSTBL(st_hull_power, tbl_shiptech_hull, power, NUMTYPE_U16, 0, 0xffff),
     DEFNUMITEMLSTBL(st_hull_defense, tbl_shiptech_hull, defense, NUMTYPE_S16, -0x8000, 0x7fff),
     DEFNUMITEML(deterministic, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(ai_trans_range_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(bt_turn_max, NUMTYPE_S, 1, 0xffff),
     DEFNUMITEML(bt_wait_no_reload, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(bt_precap_tohit, NUMTYPE_BOOL, 0, 1),
@@ -179,11 +180,16 @@ static const struct numtbl_s {
     DEFNUMITEML(max_ship_maint, NUMTYPE_S, 1, 0x7fffffff),
     DEFNUMITEML(max_tribute_bc, NUMTYPE_S, 1, 0x7fffffff),
     DEFNUMITEML(event_roll, NUMTYPE_S, 0, 0x7fffffff),
+    DEFNUMITEML(accident_chk_factories, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(council_years, NUMTYPE_S, 0, 0x7fff),
     DEFNUMITEML(news_orion, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(aud_ask_break_nap, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(aud_bounty_give, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(aud_update_tech, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(update_on_load, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(newtech_adjust_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(monster_rest_att, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(orbital_bio_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(orbital_weap_any, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(orbital_weap_4, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(orbital_torpedo, NUMTYPE_BOOL, 0, 1),
@@ -195,8 +201,21 @@ static const struct numtbl_s {
     DEFNUMITEML(first_tech_rp_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(waste_calc_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(waste_adjust_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(slider_respects_locks, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(pop_tenths_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(factory_cost_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(doom_stack_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEML(eco_slider_slack, NUMTYPE_U8, 0, 0xff),
+    DEFNUMITEML(reset_tform_to_max, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(soil_rounding_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(leaving_trans_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(hidden_child_labor_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(extended_reloc_range, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(colonized_factories_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(cond_switch_to_ind_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(slider_eco_done_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(ship_scanner_fix, NUMTYPE_BOOL, 0, 1),
+    DEFNUMITEML(ai_fleet_cheating_fix, NUMTYPE_BOOL, 0, 1),
     DEFNUMITEMLTBLL(tbl_hull_w, NUMTYPE_U8, 0, 0xff),
     DEFNUMITEMLTBLL(tech_costmuld, NUMTYPE_U8, 0, 0xff),
     DEFNUMITEMLTBLL(tech_costmula, NUMTYPE_U8, 0, 0xff),
@@ -434,13 +453,53 @@ static const struct numtbl_s *find_match(const char *numid, int i, int num)
 
 /* -------------------------------------------------------------------------- */
 
+bool game_opt_fix_bugs = false;
+bool game_opt_fix_guardian_repair = false;
+bool game_opt_fix_starting_ships = false;
+
+void game_num_fix_bugs(void)
+{
+    game_num_ng_tech[RACE_SILICOID][TECH_FIELD_PLANETOLOGY][24] = 1;
+    game_num_ng_tech[RACE_SILICOID][TECH_FIELD_PLANETOLOGY][30] = 1;
+    game_num_max_factories = 2700;
+    game_num_accident_chk_factories = true;
+    game_num_weapon_list_max = 64;
+    game_num_aud_bounty_give = true;
+    game_num_aud_update_tech = true;
+    game_num_update_on_load = false;
+    game_num_combat_trans_fix = true;
+    game_num_eco_slider_slack = 0;
+    game_num_reset_tform_to_max = false;
+}
+
+void game_num_fix_guardian_repair(void)
+{
+    tbl_monster[MONSTER_GUARDIAN][3].special[1] = 16;
+    tbl_monster[MONSTER_GUARDIAN][4].special[1] = 26;
+    tbl_monster[MONSTER_GUARDIAN][4].repair = 30;
+}
+
+void game_num_fix_starting_ships(void)
+{
+    tbl_startship[0].engines = 2;
+    tbl_startship[1].engines = 27;
+    tbl_startship[2].engines = 110;
+    tbl_startship[3].engines = 85;
+    tbl_startship[4].engines = 100;
+    tbl_startship[0].cost = 8;
+    tbl_startship[1].cost = 14;
+    tbl_startship[2].cost = 74;
+    tbl_startship[3].cost = 65;
+    tbl_startship[4].cost = 570;
+}
+
 bool game_num_patch(const char *numid, const int32_t *patchnums, int first, int num)
 {
     const struct numtbl_s *s = find_match(numid, first, num);
     if (s) {
         void *p;
         bool is_signed;
-        p = s->ptr + first * s->tstep;
+        p = ((uint8_t *)s->ptr) + first * s->tstep;
         is_signed = NUMTYPE_IS_S(s->numtype);
         for (int i = 0; i < num; ++i) {
             int32_t v;
@@ -464,7 +523,7 @@ bool game_num_patch(const char *numid, const int32_t *patchnums, int first, int
                 case NUMTYPE_U32: *((uint32_t *)p) = (uint32_t)v; break;
                 case NUMTYPE_BOOL: *((bool *)p) = (v != 0); break;
             }
-            p += s->tstep;
+            p = ((uint8_t *)p) + s->tstep;
         }
         return true;
     } else {
@@ -506,7 +565,7 @@ void game_num_dump(void)
             } else {
                 log_message(",%u", vu);
             }
-            p += s->tstep;
+            p = ((uint8_t *)p) + s->tstep;
         }
         log_message("\n");
         ++s;
