@@ -12,7 +12,8 @@
 #include "hwsdl_opt.h"
 #include "lib.h"
 #include "log.h"
-#include "screenshot.h"
+#include "mouse.h"
+#include "palette.h"
 #include "types.h"
 #include "version.h"
 
@@ -60,12 +61,14 @@ static struct sdl_video_s {
     int w_upscale, h_upscale;
     int actualh;
 
+    bool noblit;
     bool need_resize;
+    bool shrink, enlarge;
     int last_resize_time;
 
     void (*render)(int bufi);
     void (*update)(void);
-    void (*setpal)(uint8_t *pal, int first, int num);
+    void (*setpal)(const uint8_t *pal, int first, int num);
 
     /* buffers used by UI */
     uint8_t *buf[NUM_VIDEOBUF];
@@ -73,18 +76,63 @@ static struct sdl_video_s {
     int bufh;
     int bufi;
 
-    /* palette as set by UI, 6bpp */
-    uint8_t pal[256 * 3];
     /* palette as used by SDL */
     SDL_Color color[256];
     bool palette_to_set;
-    bool flag_screenshot;
 } video = { 0 };
 
 /* -------------------------------------------------------------------------- */
 
+#include "hwsdl_video.c"
+
+/* -------------------------------------------------------------------------- */
+
+static void video_create_texture(void)
+{
+    const char *scaling_quality_str = NULL;
+    if (video.texture != NULL) {
+        SDL_DestroyTexture(video.texture);
+    }
+    /* Set the scaling quality for rendering the intermediate texture into
+       the upscaled texture to "nearest" (by default), which is gritty and
+       pixelated and resembles software scaling pretty well.
+    */
+    switch (hw_opt_scaling_quality) {
+        case 1:
+            scaling_quality_str = "linear";
+            break;
+        case 2:
+            scaling_quality_str = "best";
+            break;
+        default:
+            scaling_quality_str = "nearest";
+            break;
+    }
+    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, scaling_quality_str);
+    /* Create the intermediate texture that the RGBA surface gets loaded into.
+       The SDL_TEXTUREACCESS_STREAMING flag means that this texture's content
+       is going to change frequently.
+    */
+    log_message("SDL_CreateTexture: source, %d, %d, %s\n",
+                video.bufw, video.bufh, scaling_quality_str);
+    video.texture = SDL_CreateTexture(video.renderer,
+                                      video.pixel_format,
+                                      SDL_TEXTUREACCESS_STREAMING,
+                                      video.bufw, video.bufh);
+}
+
 static void video_create_upscaled_texture(bool force)
 {
+    if (!hw_opt_allow_upscaling || hw_opt_int_scaling) {
+        if (video.texture_upscaled) {
+            SDL_DestroyTexture(video.texture_upscaled);
+            video.texture_upscaled = NULL;
+            video.w_upscale = 0;
+            video.h_upscale = 0;
+        }
+        return;
+    }
+
     int w, h;
     int h_upscale, w_upscale;
 
@@ -108,6 +156,16 @@ static void video_create_upscaled_texture(bool force)
         w = (h * video.bufw) / video.actualh;
     }
 
+    if ((w % video.bufw == 0) && (h % video.bufh == 0)) {
+        if (video.texture_upscaled) {
+            SDL_DestroyTexture(video.texture_upscaled);
+            video.texture_upscaled = NULL;
+            video.w_upscale = 0;
+            video.h_upscale = 0;
+        }
+        return;
+    }
+
     /* Pick texture size the next integer multiple of the screen dimensions.
        If one screen dimension matches an integer multiple of the original
        resolution, there is no need to overscale in this direction.
@@ -138,6 +196,8 @@ static void video_create_upscaled_texture(bool force)
     */
     SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");
 
+    log_message("SDL_CreateTexture: upscaled, %d, %d\n",
+                w_upscale * video.bufw, h_upscale * video.bufh);
     video.texture_upscaled = SDL_CreateTexture(video.renderer,
                                 video.pixel_format,
                                 SDL_TEXTUREACCESS_TARGET,
@@ -147,10 +207,26 @@ static void video_create_upscaled_texture(bool force)
 
 }
 
+static void video_destroy_renderer(void)
+{
+    if (video.renderer) {
+        SDL_DestroyRenderer(video.renderer);
+        // all associated textures get destroyed
+        video.renderer = NULL;
+        video.texture = NULL;
+        video.texture_upscaled = NULL;
+        video.w_upscale = 0;
+        video.h_upscale = 0;
+    }
+}
+
 /* -------------------------------------------------------------------------- */
 
 static void video_render(int bufi)
 {
+    if (video.noblit) {
+        return;
+    }
     int pitch = video.screen->pitch;
     Uint8 *p = (Uint8 *)video.screen->pixels;
     uint8_t *q = video.buf[bufi];
@@ -174,6 +250,35 @@ static void video_adjust_window_size(int *wptr, int *hptr)
         } else {
             w = (h * video.bufw) / video.actualh;
         }
+        bool do_resize = false;
+        int scale = w / video.bufw;
+        if (video.shrink || video.enlarge) {
+            if (video.shrink) {
+                --scale;
+                if (w % video.bufw) {
+                    ++scale;
+                }
+            } else {
+                ++scale;
+            }
+            do_resize = true;
+        } else if (hw_opt_int_scaling) {
+            if (video_check_opt_screen_winwh() == 0) {
+                if ((hw_opt_screen_winw < *wptr) && (hw_opt_screen_winh <= *hptr)) {
+                    ++scale;
+                } else if ((hw_opt_screen_winw <= *wptr) && (hw_opt_screen_winh < *hptr)) {
+                    ++scale;
+                }
+            }
+            do_resize = true;
+        }
+        if (do_resize) {
+            if (scale <= 0) {
+                scale = 1;
+            }
+            w = video.bufw * scale;
+            h = video.actualh * scale;
+        }
         *wptr = w;
         *hptr = h;
     }
@@ -181,6 +286,9 @@ static void video_adjust_window_size(int *wptr, int *hptr)
 
 static void video_update(void)
 {
+    if (video.noblit) {
+        return;
+    }
     if (video.need_resize) {
         if (SDL_GetTicks() > (video.last_resize_time + RESIZE_DELAY)) {
             int flags, w, h;
@@ -188,16 +296,21 @@ static void video_update(void)
                save the new window size.
             */
             flags = SDL_GetWindowFlags(video.window);
+            SDL_GetWindowSize(video.window, &w, &h);
             if ((flags & (SDL_WINDOW_FULLSCREEN_DESKTOP | SDL_WINDOW_FULLSCREEN | SDL_WINDOW_MAXIMIZED)) == 0) {
-                SDL_GetWindowSize(video.window, &w, &h);
                 /* Adjust the window by resizing again so that the window is the right aspect ratio. */
-                video_adjust_window_size(&w, &h);
+                if (hw_opt_autotrim || video.shrink || video.enlarge) {
+                    video_adjust_window_size(&w, &h);
+                }
                 SDL_SetWindowSize(video.window, w, h);
                 hw_opt_screen_winw = w;
                 hw_opt_screen_winh = h;
             }
+            hw_mouse_set_scale(w, h);
             video_create_upscaled_texture(false);
             video.need_resize = false;
+            video.shrink = false;
+            video.enlarge = false;
             video.palette_to_set = true;
         } else {
             return;
@@ -215,24 +328,20 @@ static void video_update(void)
     SDL_BlitSurface(video.screen, &video.blit_rect, video.interbuffer, &video.blit_rect);
 
     /* Update the intermediate texture with the contents of the intermediate buffer.*/
-    {
-        void *pixels;
-        int pitch;
-
-        if (SDL_LockTexture(video.texture, NULL, &pixels, &pitch) == 0) {
-            memcpy(pixels, video.interbuffer->pixels, pitch*video.bufh);
-            SDL_UnlockTexture(video.texture);
-        } else {
-            log_error("SDL_LockTexture(video.texture): %s\n", SDL_GetError());
-            exit(EXIT_FAILURE);
-        }
-    }
+    SDL_UpdateTexture(video.texture, NULL, video.interbuffer->pixels, video.interbuffer->pitch);
 
     /* Make sure the pillarboxes are kept clear each frame. */
     SDL_RenderClear(video.renderer);
 
+    if (video.texture_upscaled == NULL) {
+        SDL_SetRenderTarget(video.renderer, NULL);
+        SDL_RenderCopy(video.renderer, video.texture, NULL, NULL);
+        SDL_RenderPresent(video.renderer);
+        return;
+    }
+
     /* Render this intermediate texture into the upscaled texture
-       using "nearest" integer scaling.
+       using hw_opt_scaling_quality integer scaling.
     */
     SDL_SetRenderTarget(video.renderer, video.texture_upscaled);
     SDL_RenderCopy(video.renderer, video.texture, NULL, NULL);
@@ -245,13 +354,12 @@ static void video_update(void)
     SDL_RenderPresent(video.renderer);
 }
 
-static void video_setpal(uint8_t *pal, int first, int num)
+static void video_setpal(const uint8_t *pal, int first, int num)
 {
-    memcpy(&video.pal[first * 3], pal, num * 3);
     for (int i = first; i < (first + num); ++i) {
-        video.color[i].r = *pal++ << 2;
-        video.color[i].g = *pal++ << 2;
-        video.color[i].b = *pal++ << 2;
+        video.color[i].r = palette_6bit_to_8bit(*pal++);
+        video.color[i].g = palette_6bit_to_8bit(*pal++);
+        video.color[i].b = palette_6bit_to_8bit(*pal++);
         video.color[i].a = 255;
     }
     video.palette_to_set = true;
@@ -260,6 +368,38 @@ static void video_setpal(uint8_t *pal, int first, int num)
 
 /* -------------------------------------------------------------------------- */
 
+void hw_video_set_visible(bool visible)
+{
+    video.noblit = !visible;
+}
+
+#if SDL_VERSION_ATLEAST(2, 0, 18)
+bool hw_video_toggle_vsync(void)
+{
+    hw_opt_vsync = !hw_opt_vsync;
+    SDL_RenderSetVSync(video.renderer, hw_opt_vsync);
+    return true;
+}
+#endif
+
+#if SDL_VERSION_ATLEAST(2, 0, 5)
+bool hw_video_toggle_int_scaling(void)
+{
+    hw_opt_int_scaling = !hw_opt_int_scaling;
+    SDL_RenderSetIntegerScale(video.renderer, hw_opt_int_scaling);
+    return true;
+}
+#endif
+
+#if SDL_VERSION_ATLEAST(2, 0, 12)
+bool hw_video_filter_next(void)
+{
+    hw_opt_scaling_quality = (hw_opt_scaling_quality + 1) % 3;
+    SDL_SetTextureScaleMode(video.texture, hw_opt_scaling_quality);
+    return true;
+}
+#endif
+
 /* -------------------------------------------------------------------------- */
 
 /* Check the display bounds of the display referred to by 'video_display' and
@@ -305,18 +445,7 @@ static void video_get_window_position(int *x, int *y, int w, int h)
 
 static void video_window_destroy(void)
 {
-    if (video.texture) {
-        SDL_DestroyTexture(video.texture);
-        video.texture = NULL;
-    }
-    if (video.texture_upscaled) {
-        SDL_DestroyTexture(video.texture_upscaled);
-        video.texture_upscaled = NULL;
-    }
-    if (video.renderer) {
-        SDL_DestroyRenderer(video.renderer);
-        video.renderer = NULL;
-    }
+    video_destroy_renderer();
     if (video.window) {
         SDL_DestroyWindow(video.window);
         video.window = NULL;
@@ -328,12 +457,16 @@ static int video_sw_set(int w, int h)
     int x, y;
     int window_flags = 0, renderer_flags = 0;
     SDL_DisplayMode mode;
+    SDL_RendererInfo info;
 
     /* In windowed mode, the window can be resized while the game is running. */
     window_flags = SDL_WINDOW_RESIZABLE;
     /* Set the highdpi flag - this makes a big difference on Macs with
        retina displays, especially when using small window sizes. */
     window_flags |= SDL_WINDOW_ALLOW_HIGHDPI;
+    if (hw_opt_borderless) {
+        window_flags |= SDL_WINDOW_BORDERLESS;
+    }
     if (hw_opt_fullscreen) {
         if (hw_opt_screen_fsw && hw_opt_screen_fsh) {
             w = hw_opt_screen_fsw;
@@ -383,21 +516,27 @@ static int video_sw_set(int w, int h)
         return -1;
     }
     /* Turn on vsync */
-    renderer_flags |= SDL_RENDERER_PRESENTVSYNC;
+    if (hw_opt_vsync) {
+        renderer_flags |= SDL_RENDERER_PRESENTVSYNC;
+    }
 
     if (hw_opt_force_sw) {
         renderer_flags |= SDL_RENDERER_SOFTWARE;
         renderer_flags &= ~SDL_RENDERER_PRESENTVSYNC;
     }
 
-    if (video.renderer) {
-        SDL_DestroyRenderer(video.renderer);
-    }
+    video_destroy_renderer();
     video.renderer = SDL_CreateRenderer(video.window, -1, renderer_flags);
     if (video.renderer == NULL) {
         log_error("SDL2: Error creating renderer for screen window: %s\n", SDL_GetError());
         return -1;
     }
+    if (!SDL_GetRendererInfo(video.renderer, &info)) {
+        log_message("SDL_GetRendererInfo: %s%s%s\n",
+                    info.name,
+                    (info.flags & SDL_RENDERER_ACCELERATED) ? ", accelerated" : "",
+                    (info.flags & SDL_RENDERER_PRESENTVSYNC) ? ", vsync" : "");
+    }
     if (hw_opt_aspect != 0) {
         /* Important: Set the "logical size" of the rendering context. At the same
            time this also defines the aspect ratio that is preserved while scaling
@@ -450,25 +589,16 @@ static int video_sw_set(int w, int h)
             return -1;
         }
     }
-    if (video.texture != NULL) {
-        SDL_DestroyTexture(video.texture);
-    }
-    /* Set the scaling quality for rendering the intermediate texture into
-       the upscaled texture to "nearest", which is gritty and pixelated and
-       resembles software scaling pretty well.
-    */
-    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "nearest");
-    /* Create the intermediate texture that the RGBA surface gets loaded into.
-       The SDL_TEXTUREACCESS_STREAMING flag means that this texture's content
-       is going to change frequently.
-    */
-    video.texture = SDL_CreateTexture(video.renderer,
-                                video.pixel_format,
-                                SDL_TEXTUREACCESS_STREAMING,
-                                video.bufw, video.bufh);
+
+    video_create_texture();
 
     /* Initially create the upscaled texture for rendering to screen */
     video_create_upscaled_texture(true);
+
+    if (hw_opt_autotrim) {
+        hw_video_resize(0, 0);
+    }
+
     return 0;
 }
 
@@ -495,6 +625,31 @@ int hw_video_resize(int w, int h)
     return 0;
 }
 
+void hw_video_shrink(void)
+{
+    video.shrink = true;
+    video.enlarge = false;
+    hw_video_resize(0, 0);
+}
+
+void hw_video_enlarge(void)
+{
+    video.shrink = false;
+    video.enlarge = true;
+    hw_video_resize(0, 0);
+}
+
+bool hw_video_toggle_borderless(void)
+{
+    hw_opt_borderless = !hw_opt_borderless;
+    video_window_destroy();
+    if (video_sw_set(hw_opt_screen_winw, hw_opt_screen_winh) != 0) {
+        hw_opt_borderless = !hw_opt_borderless; /* restore the setting for the config file */
+        return false;
+    }
+    return true;
+}
+
 bool hw_video_toggle_fullscreen(void)
 {
     hw_opt_fullscreen = !hw_opt_fullscreen;
@@ -521,7 +676,6 @@ int hw_video_init(int w, int h)
     hw_mouse_set_limits(w, h);
     video.bufw = w;
     video.bufh = h;
-    video.flag_screenshot = false;
     video.window = NULL;
     video.renderer = NULL;
     video.screen = NULL;
@@ -531,7 +685,10 @@ int hw_video_init(int w, int h)
     video.display = 0;
     video.w_upscale = 0;
     video.h_upscale = 0;
+    video.noblit = false;
     video.need_resize = false;
+    video.shrink = false;
+    video.enlarge = false;
     video.last_resize_time = 0;
     video.render = video_render;
     video.update = video_update;
@@ -542,23 +699,46 @@ int hw_video_init(int w, int h)
     video.blit_rect.h = video.bufh;
     hw_video_update_actual_h();
     SETMAX(h, video.actualh);
-    if ((hw_opt_screen_winw != 0) && (hw_opt_screen_winh != 0)) {
+    if (video_check_opt_screen_winwh() == 0) {
         w = hw_opt_screen_winw;
         h = hw_opt_screen_winh;
-        if ((w < video.bufw) || (h < video.bufh)) {
-            log_warning("ignoring too small configured resolution %ix%i < %ix%i\n", w, h, video.bufw, video.bufh);
-            w = video.bufw;
-            h = video.bufh;
+    } else {
+        SDL_Rect bounds;
+        if (SDL_GetDisplayUsableBounds(video.display, &bounds) != 0) {
+            log_error("SDL2: Could not get display usable bounds for video display #%d: %s\n", video.display, SDL_GetError());
+            return -1;
+        }
+        int maxw = bounds.w;
+        int maxh = bounds.h;
+        {
+            int top, left, bottom, right;
+            if (SDL_GetWindowBordersSize(video.window, &top, &left, &bottom, &right) == 0) {
+                maxw -= (left + right);
+                maxh -= (top + bottom);
+            } else {
+                maxw -= 50;
+                maxh -= 50;
+            }
+        }
+        {
+            int scale_w = maxw / video.bufw;
+            int scale_h = maxh / video.actualh;
+            int scale = (scale_w > scale_h) ? scale_h : scale_w;
+            if (scale <= 0) {
+                scale = 1;
+            }
+            w = video.bufw * scale;
+            h = video.actualh * scale;
         }
     }
     if (video_sw_set(w, h)) {
         return -1;
     }
     for (int i = 0; i < NUM_VIDEOBUF; ++i) {
-        video.buf[i] = lib_malloc(w * h);
+        video.buf[i] = lib_malloc(video.bufw * video.bufh);
     }
     video.bufi = 0;
-    memset(video.pal, 0, sizeof(video.pal));
+    ui_palette_clear();
     hw_video_refresh_palette();
     return 0;
 }
@@ -593,9 +773,23 @@ void hw_video_input_grab(bool grab)
     SDL_SetWindowGrab(video.window, grab);
     if (hw_opt_relmouse) {
         SDL_SetRelativeMouseMode(grab);
+    } else {
+        hw_mouse_set_xy(moouse_x, moouse_y);
+        SDL_ShowCursor(grab ? SDL_DISABLE : SDL_ENABLE);
     }
 }
 
+void hw_video_mouse_warp(int mx, int my)
+{
+#if SDL_VERSION_ATLEAST(2, 0, 18)
+    int x, y;
+    SDL_RenderLogicalToWindow(video.renderer, mx, my, &x, &y);
+    SDL_WarpMouseInWindow(video.window, x, y);
+#else
+    mouse_set_xy_from_hw(mx, my);
+#endif
+}
+
 int hw_icon_set(const uint8_t *data, const uint8_t *pal, int w, int h)
 {
     SDL_Color color[256];
@@ -614,9 +808,9 @@ int hw_icon_set(const uint8_t *data, const uint8_t *pal, int w, int h)
         p += icon->pitch;
     }
     for (int i = 0; i < 256; ++i) {
-        color[i].r = *pal++ << 2;
-        color[i].g = *pal++ << 2;
-        color[i].b = *pal++ << 2;
+        color[i].r = palette_6bit_to_8bit(*pal++);
+        color[i].g = palette_6bit_to_8bit(*pal++);
+        color[i].b = palette_6bit_to_8bit(*pal++);
         color[i].a = 255;
     }
     color[0].a = 0;
@@ -645,5 +839,3 @@ int hw_icon_set(const uint8_t *data, const uint8_t *pal, int w, int h)
     video.icon = icon;
     return 0;
 }
-
-#include "hwsdl_video.c"
