@@ -7,6 +7,7 @@
 #include "bits.h"
 #include "cfg.h"
 #include "comp.h"
+#include "font8x8_draw.h"
 #include "game_shiptech.h"  /* for sounds used */
 #include "gfxaux.h"
 #include "hw.h"
@@ -18,9 +19,11 @@
 #include "lib.h"
 #include "log.h"
 #include "options.h"
+#include "palette.h"
 #include "types.h"
 #include "uicursor.h"
 #include "uidefs.h"
+#include "uidelay.h"
 #include "uiobj.h"
 #include "uipal.h"
 #include "uiobj.h"
@@ -32,13 +35,31 @@ static int ui_icon = 146/*guardian*/;
 
 /* -------------------------------------------------------------------------- */
 
+void ui_extra_toggle_preset(bool enabled)
+{
+    ui_extra_enabled = enabled;
+    ui_fixbugs_enabled = enabled;
+    ui_load_opts_extra = enabled;
+    game_opt_skip_intro_always = enabled;
+    ui_space_combat_autoresolve = enabled;
+    ui_sm_ships_enabled = enabled;
+    ui_sm_distance_tooltip = enabled;
+    ui_sm_expanded_scroll = enabled;
+    ui_sm_explicit_cursor_context = enabled;
+    ui_sm_no_question_mark_cursor = enabled;
+    ui_illogical_hotkey_fix = enabled;
+    ui_kbd_cursor_keys_fix = enabled;
+    ui_mouse_lmb_fix = enabled;
+    ui_mouse_warp_disabled = enabled;
+}
+
 static bool check_ui_scale(void *var)
 {
     int v = (int)(intptr_t)var;
-    if ((v > 0) && (v < UI_SCALE_MAX)) {
+    if ((v > 0) && (v <= UI_SCALE_MAX)) {
         return true;
     } else {
-        log_error("invalid ui_scale %i, must be 0 < N < %i\n", v, UI_SCALE_MAX);
+        log_error("invalid ui_scale %i, must be 0 < N <= %i\n", v, UI_SCALE_MAX);
         return false;
     }
 }
@@ -66,8 +87,23 @@ static bool check_ui_sm_scroll_speed(void *var)
 }
 
 const struct cfg_items_s ui_cfg_items[] = {
-    CFG_ITEM_INT("uiscale", &ui_scale, check_ui_scale),
+    CFG_ITEM_INT("uiscale", &ui_scale_hint, check_ui_scale),
     CFG_ITEM_BOOL("uiextra", &ui_extra_enabled),
+    CFG_ITEM_BOOL("uifixbugs", &ui_fixbugs_enabled),
+    CFG_ITEM_BOOL("illogical_hotkey_fix", &ui_illogical_hotkey_fix),
+    CFG_ITEM_BOOL("load_opts_extra", &ui_load_opts_extra),
+    CFG_ITEM_BOOL("space_combat_autoresolve", &ui_space_combat_autoresolve),
+    CFG_ITEM_BOOL("ui_sm_ships_enabled", &ui_sm_ships_enabled),
+    CFG_ITEM_BOOL("sm_expanded_scroll", &ui_sm_expanded_scroll),
+    CFG_ITEM_BOOL("sm_distance_tooltip", &ui_sm_distance_tooltip),
+    CFG_ITEM_BOOL("sm_no_question_mark_cursor", &ui_sm_no_question_mark_cursor),
+    CFG_ITEM_BOOL("sm_explicit_cursor_context", &ui_sm_explicit_cursor_context),
+    CFG_ITEM_BOOL("mouse_lmb_fix", &ui_mouse_lmb_fix),
+    CFG_ITEM_BOOL("mouse_warp_disabled", &ui_mouse_warp_disabled),
+    CFG_ITEM_BOOL("copy_protection_disabled", &ui_copyprotection_disabled),
+    CFG_ITEM_BOOL("sm_mouseover_focus", &ui_sm_mouseover_focus),
+    CFG_ITEM_BOOL("sm_mouse_scroll", &ui_sm_mouse_scroll),
+    CFG_ITEM_BOOL("kbd_cursor_keys_fix", &ui_kbd_cursor_keys_fix),
     CFG_ITEM_COMMENT("0..146"),
     CFG_ITEM_INT("uiicon", &ui_icon, check_ui_icon),
     CFG_ITEM_COMMENT("Invert mouse wheel for sliders"),
@@ -76,6 +112,7 @@ const struct cfg_items_s ui_cfg_items[] = {
     CFG_ITEM_BOOL("mwicounter", &ui_mwi_counter),
     CFG_ITEM_COMMENT("Starmap scroll speed (1..10, 0 = instant)"),
     CFG_ITEM_INT("sm_scroll_speed", &ui_sm_scroll_speed, check_ui_sm_scroll_speed),
+    CFG_ITEM_BOOL("kbd_repeat", &ui_kbd_repeat),
     CFG_ITEM_END
 };
 
@@ -83,7 +120,7 @@ static int ui_options_set_scale(char **argv, void *var)
 {
     int v = atoi(argv[1]);
     if (check_ui_scale((void *)(intptr_t)v)) {
-        ui_scale = v;
+        ui_scale_hint = v;
         return 0;
     }
     return -1;
@@ -136,11 +173,28 @@ struct ui_data_s ui_data = { 0 };
 int ui_screen_w = 0;
 int ui_screen_h = 0;
 int ui_scale = 0;
+int ui_scale_hint = 0;
 int starmap_scale = 0;
 bool ui_extra_enabled = false;
+bool ui_fixbugs_enabled = false;
+bool ui_illogical_hotkey_fix = false;
+bool ui_load_opts_extra = false;
+bool ui_space_combat_autoresolve = false;
+bool ui_sm_ships_enabled = false;
+bool ui_sm_expanded_scroll = false;
+bool ui_sm_distance_tooltip = false;
+bool ui_sm_no_question_mark_cursor = false;
+bool ui_sm_explicit_cursor_context = false;
+bool ui_mouse_lmb_fix = false;
+bool ui_mouse_warp_disabled = false;
+bool ui_copyprotection_disabled = false;
+bool ui_sm_mouseover_focus = false;
+bool ui_sm_mouse_scroll = false;
+bool ui_kbd_cursor_keys_fix = false;
 bool ui_mwi_slider = false;
 bool ui_mwi_counter = false;
 int ui_sm_scroll_speed = 3;
+bool ui_kbd_repeat = true;
 
 bool ui_use_audio = true;
 
@@ -349,7 +403,6 @@ static void init_gfx(void)
     gfx_aux_setup_wh(&ui_data.aux.ship_overlay, 34, 26);
     gfx_aux_setup_wh(&ui_data.aux.btemp, 50, 40);
 
-    memset(ui_data.starmap.tag, PLANET_NONE, sizeof(ui_data.starmap.tag));
     ui_data.gfx.initialized = true;
 }
 
@@ -389,6 +442,36 @@ static int set_ui_icon(void)
 
 /* -------------------------------------------------------------------------- */
 
+void ui_early_show_message_box(const char *msg)
+{
+    if (hw_video_init(320, 200)) {
+        return;
+    }
+    bool initialized = false;
+    int64_t init_time = hw_get_time_us();
+    ui_palette_clear();
+    ui_palette_set_color(0, 0, 0, 0);
+    ui_palette_set_color(1, 0x3f, 0x3f, 0x3f);
+    hw_video_refresh_palette();
+    while (1) {
+        if (!initialized && (hw_get_time_us() - init_time) > 1000000) {
+            memset(hw_video_get_buf(), 0, 320 * 200);
+            font8x8_drawstr_rect(50, 20, 270, 180, 320, msg, 1, 0);
+            hw_video_draw_buf();
+            initialized = true;
+        }
+        hw_event_handle();
+        if (kbd_have_keypress()) {
+            uint32_t k;
+            k = kbd_get_keypress();
+            if ((KBD_GET_KEY(k) == MOO_KEY_SPACE)
+              ||(KBD_GET_KEY(k) == MOO_KEY_ESCAPE)) {
+                break;
+            }
+        }
+    }
+}
+
 int ui_early_init(void)
 {
     return 0;
@@ -406,10 +489,11 @@ int ui_late_init(void)
         log_error("V11.LBX not found! Make sure that your MOO1 is updated to v1.3.\n");
         return 1;
     }
-    if (ui_scale == 0) {
-        ui_scale = 1;
+    if (ui_scale_hint == 0) {
+        ui_scale_hint = 1;
     }
-    starmap_scale = 1;
+    ui_scale = ui_scale_hint;
+    starmap_scale = ui_scale_hint;
     ui_screen_w = UI_VGA_W * ui_scale;
     ui_screen_h = UI_VGA_H * ui_scale;
     ui_cursor_init(ui_scale);
@@ -479,6 +563,7 @@ int ui_late_init(void)
     }
     uiobj_table_clear();
     kbd_clear();
+    hw_kbd_set_repeat(ui_kbd_repeat);
     return 0;
 }
 
