@@ -102,13 +102,13 @@ static void newtech_draw_cb1(void *vptr)
             lbxgfx_set_new_frame(d->gfx_spies, frame);
         }
     }
-    game_tech_get_newtech_msg(g, d->api, &(d->nt), buf);
+    game_tech_get_newtech_msg(g, d->api, &(d->nt), buf, UI_STRBUF_SIZE);
     lbxfont_select_set_12_4(5, 5, 0, 0);
     lbxfont_print_str_center(161, 7, buf, UI_SCREEN_W, ui_scale);
     if (d->nt.source != TECHSOURCE_CHOOSE) {
         int strh, y;
         char *p, c;
-        game_tech_get_name(d->g->gaux, d->nt.field, d->nt.tech, buf);
+        game_tech_get_name(d->g->gaux, d->nt.field, d->nt.tech, buf, UI_STRBUF_SIZE);
         p = buf;
         while ((c = *p) != 0) {
             if (islower(c)) {
@@ -118,7 +118,7 @@ static void newtech_draw_cb1(void *vptr)
         }
         lbxfont_select_set_12_4(5, 8, 0, 0);
         lbxfont_print_str_center(160, 32, buf, UI_SCREEN_W, ui_scale);
-        game_tech_get_descr(d->g->gaux, d->nt.field, d->nt.tech, buf);
+        game_tech_get_descr(d->g->gaux, d->nt.field, d->nt.tech, buf, UI_STRBUF_SIZE);
         lbxfont_select_set_12_5(4, 0xf, 0, 0);
         strh = lbxfont_calc_split_str_h(305, buf);
         /* BUG?
@@ -166,13 +166,13 @@ static void newtech_choose_next_draw_cb(void *vptr)
        On DOS/v1.3 this only overwrite unused VRAM. We must the the _offs version.
     */
     lbxgfx_draw_frame_offs(x, y + yo, d->gfx_pulldown_d, 0, 0, UI_VGA_W - 1, UI_VGA_H - 1, UI_SCREEN_W, ui_scale);
-    sprintf(buf, "%s %s", game_str_tbl_te_field[d->nt.field], game_str_te_techno);
+    lib_sprintf(buf, sizeof(buf), "%s %s", game_str_tbl_te_field[d->nt.field], game_str_te_techno);
     lbxfont_select(5, 0xe, 0, 0);
     lbxfont_print_str_center(x + 85, y + 5, buf, UI_SCREEN_W, ui_scale);
     lbxfont_select_set_12_1(0, 0, 0xe, 0);
-    game_tech_get_descr(d->g->gaux, d->nt.field, tech, buf);
+    game_tech_get_descr(d->g->gaux, d->nt.field, tech, buf, sizeof(buf));
     pos = strlen(buf);
-    sprintf(&buf[pos], " \x02(%u %s)\x1", game_tech_get_next_rp(d->g, d->api, d->nt.field, tech), game_str_te_rp);
+    lib_sprintf(&buf[pos], sizeof(buf) - pos, " \x02(%u %s)\x1", game_tech_get_next_rp(d->g, d->api, d->nt.field, tech), game_str_te_rp);
     lbxfont_print_str_split(151, y + yo + 18, 156, buf, 0, UI_SCREEN_W, UI_SCREEN_H, ui_scale);
 }
 
@@ -189,7 +189,7 @@ static void ui_newtech_choose_next(struct newtech_data_s *d)
         uint8_t tech;
         tech = d->tech_next[i];
         d->tbl_tech[i] = tech; /* WASBUG overwrites flag_copybuf, flag_dialog and d->tech_next */
-        game_tech_get_name(d->g->gaux, d->nt.field, tech, tname[i]);
+        game_tech_get_name(d->g->gaux, d->nt.field, tech, tname[i], 35);
         nptr[i] = tname[i];
     }
     nptr[d->num_next] = 0;
@@ -215,12 +215,12 @@ static void ui_newtech_choose_next(struct newtech_data_s *d)
     uiobj_unset_callback();
 }
 
-static void newtech_adjust_draw_typestr(char *buf, const char *str1, const char *str2, int x, int y, int we)
+static void newtech_adjust_draw_typestr(char *buf, size_t bufsize, const char *str1, const char *str2, int x, int y, int we)
 {
-    strcat(buf, game_str_nt_inc);
-    strcat(buf, str1);
+    lib_strcat(buf, game_str_nt_inc, bufsize);
+    lib_strcat(buf, str1, bufsize);
     if (str2) {
-        strcat(buf, str2);
+        lib_strcat(buf, str2, bufsize);
     }
     lbxfont_print_str_split(x + 15, y + 16, 110 + we, buf, 3, UI_SCREEN_W, UI_SCREEN_H, ui_scale);
 }
@@ -234,10 +234,10 @@ static void newtech_adjust_draw_cb(void *vptr)
     ui_draw_filled_rect(x, y, x + 135, y + 80, 0xf9, ui_scale);
     lbxgfx_draw_frame(x, y, (d->dialog_type == 0) ? d->gfx_eco_chng2 : d->gfx_eco_chng4, UI_SCREEN_W, ui_scale);
     lbxfont_select_set_12_1(0, 0, 0, 0);
-    strcpy(buf, game_str_nt_doyou);
+    lib_strcpy(buf, game_str_nt_doyou, sizeof(buf));
     switch (d->dialog_type) {
         case 0:
-            strcat(buf, game_str_nt_redueco);
+            lib_strcat(buf, game_str_nt_redueco, sizeof(buf));
             lbxfont_print_str_split(x + 15, y + 11, 110, buf, 3, UI_SCREEN_W, UI_SCREEN_H, ui_scale);
             lbxgfx_set_frame_0(ui_data.gfx.starmap.scrapbut_yes);
             lbxgfx_set_frame_0(ui_data.gfx.starmap.scrapbut_no);
@@ -245,19 +245,19 @@ static void newtech_adjust_draw_cb(void *vptr)
             lbxgfx_draw_frame(x + 18, y + 60, ui_data.gfx.starmap.scrapbut_no, UI_SCREEN_W, ui_scale);
             break;
         case 1:
-            newtech_adjust_draw_typestr(buf, game_str_nt_ind, 0, x, y, 0);
+            newtech_adjust_draw_typestr(buf, sizeof(buf), game_str_nt_ind, 0, x, y, 0);
             break;
         case 2:
-            newtech_adjust_draw_typestr(buf, game_str_nt_ecoall, game_str_nt_terra, x, y, 0);
+            newtech_adjust_draw_typestr(buf, sizeof(buf), game_str_nt_ecoall, game_str_nt_terra, x, y, 0);
             break;
         case 3:
-            newtech_adjust_draw_typestr(buf, game_str_nt_def, 0, x, y - 5, 0);
+            newtech_adjust_draw_typestr(buf, sizeof(buf), game_str_nt_def, 0, x, y - 5, 0);
             break;
         case 4:
-            newtech_adjust_draw_typestr(buf, game_str_nt_ecostd, game_str_nt_terra, x - 1, y - 5, 2);
+            newtech_adjust_draw_typestr(buf, sizeof(buf), game_str_nt_ecostd, game_str_nt_terra, x - 1, y - 5, 2);
             break;
         case 5:
-            newtech_adjust_draw_typestr(buf, game_str_nt_ecohost, game_str_nt_terra, x - 1, y - 5, 2);
+            newtech_adjust_draw_typestr(buf, sizeof(buf), game_str_nt_ecohost, game_str_nt_terra, x - 1, y - 5, 2);
             break;
         default:
             break;
@@ -305,7 +305,9 @@ static void ui_newtech_adjust(struct newtech_data_s *d)
                 const int tbl_gr[5] = { 0, 0, 0, 1, 2 };
                 ui_sound_play_sfx_24();
                 game_update_tech_util(d->g);
-                game_update_eco_on_waste(d->g, d->api, true);
+                if (!game_num_newtech_adjust_fix) {
+                    game_update_eco_on_waste(d->g, d->api, true);
+                }
                 game_planet_adjust_percent(d->g, d->api, tbl_si[d->dialog_type - 1], game_num_tbl_tech_autoadj[i + 1], tbl_gr[d->dialog_type - 1]);
                 flag_done = true;
             }
@@ -484,9 +486,7 @@ void ui_newtech(struct game_s *g, int pi)
         {
             int v;
             if (d.nt.tech <= 50) {
-                const uint8_t *p;
-                p = RESEARCH_D0_PTR(g->gaux, d.nt.field, d.nt.tech);
-                v = p[2];
+                v = game_tech_get_gfx_i(g->gaux, d.nt.field, d.nt.tech);
             } else {
                 const int tbl[TECH_FIELD_NUM] = { 9, 22, 20, 25, 15, 29 };
                 v = tbl[d.nt.field];
