@@ -61,7 +61,7 @@ static void free_pl_data(struct planets_data_s *d)
     lbxfile_item_release(LBXFILE_BACKGRND, d->gfx_transfer);
 }
 
-static const char *planets_get_notes_str(const struct game_s *g, uint8_t pli, bool *flag_normal_ptr, char *buf)
+static const char *planets_get_notes_str(const struct game_s *g, uint8_t pli, bool *flag_normal_ptr, char *buf, size_t bufsize)
 {
     const planet_t *p = &(g->planet[pli]);
     const char *str = NULL;
@@ -89,8 +89,8 @@ static const char *planets_get_notes_str(const struct game_s *g, uint8_t pli, bo
         }
         if (p->have_stargate) {
             if (str) {
-                strcpy(buf, str);
-                strcat(buf, " *");
+                lib_strcpy(buf, str, bufsize);
+                lib_strcat(buf, " *", bufsize);
                 str = buf;
              } else {
                  str = game_str_sm_stargate;
@@ -103,6 +103,17 @@ static const char *planets_get_notes_str(const struct game_s *g, uint8_t pli, bo
     return str;
 }
 
+static const char *planets_get_dock_str(const struct game_s *g, const struct planet_s *p)
+{
+    const char *str;
+    if (p->buildship == BUILDSHIP_STARGATE) {
+        str = game_str_sm_stargate;
+    } else {
+        str = g->srd[p->owner].design[p->buildship].name;
+    }
+    return str;
+}
+
 static void planets_draw_cb(void *vptr)
 {
     struct planets_data_s *d = vptr;
@@ -136,7 +147,7 @@ static void planets_draw_cb(void *vptr)
     lbxfont_print_str_normal(210, 170, game_str_pl_reserve, UI_SCREEN_W, ui_scale);
     lbxfont_select(2, 6, 0, 0);
     v = (e->tax * e->total_production_bc) / 2000;
-    sprintf(buf, "+%i", v);
+    lib_sprintf(buf, sizeof(buf), "+%i", v);
     lbxfont_print_str_right(272, 159, buf, UI_SCREEN_W, ui_scale);
     lbxfont_print_str_normal(276, 159, game_str_bc, UI_SCREEN_W, ui_scale);
     lbxfont_print_str_normal(276, 170, game_str_bc, UI_SCREEN_W, ui_scale);
@@ -155,15 +166,8 @@ static void planets_draw_cb(void *vptr)
             p = &(g->planet[pli]);
             lbxfont_select(2, 0xb, 0, 0);
             lbxfont_print_num_right(17, y0, ui_data.sorted.index[pi] + 1, UI_SCREEN_W, ui_scale);
-            lbxfont_select(2, BOOLVEC_IS1(p->extras, PLANET_EXTRAS_GOVERNOR) ? 0xe : 0xd, 0, 0);
+            lbxfont_select(2, 0xd, 0, 0);
             lbxfont_print_str_normal(25, y0, p->name, UI_SCREEN_W, ui_scale);
-            {
-                /* draw box to highlight governor; use a color based on mode */
-                uint8_t c = ui_draw_govern_color(p, d->api);
-                if (c != 0) {
-                    ui_draw_filled_rect(25 + 40, y0 + 1, 25 + 40 + 1, y0 + 4, c, ui_scale);
-                }
-            }
             lbxfont_select(2, 6, 0, 0);
             lbxfont_print_num_right(83, y0, p->pop, UI_SCREEN_W, ui_scale);
             if (p->pop != p->pop_prev) {
@@ -180,11 +184,16 @@ static void planets_draw_cb(void *vptr)
                     gfx = ui_data.gfx.starmap.gr_arrow_d;
                 }
                 lbxgfx_draw_frame(92, y0 - 1, gfx, UI_SCREEN_W, ui_scale);
-                sprintf(buf, "%c%i", c, v);
+                lib_sprintf(buf, sizeof(buf), "%c%i", c, v);
                 lbxfont_print_str_right(111, y0, buf, UI_SCREEN_W, ui_scale);
             }
-            lbxfont_print_str_right(149, y0, game_str_tbl_roman[p->shield], UI_SCREEN_W, ui_scale);
-            /*lbxfont_select(2, 6, 0, 0);*/
+            if (ui_extra_enabled && p->battlebg == 0) {
+                lbxfont_select(2, 1, 0, 0);
+                lbxfont_print_str_right(149, y0, "neb", UI_SCREEN_W, ui_scale);
+            } else {
+                lbxfont_print_str_right(149, y0, game_str_tbl_roman[p->shield], UI_SCREEN_W, ui_scale);
+            }
+            lbxfont_select(2, 6, 0, 0);
             lbxfont_print_num_right(132, y0, p->factories, UI_SCREEN_W, ui_scale);
             lbxfont_print_num_right(170, y0, p->missile_bases, UI_SCREEN_W, ui_scale);
             if (p->waste) {
@@ -200,7 +209,7 @@ static void planets_draw_cb(void *vptr)
             lbxfont_select(2, 0xb, 0, 0);
             {
                 bool flag_normal;
-                str = planets_get_notes_str(g, pli, &flag_normal, buf);
+                str = planets_get_notes_str(g, pli, &flag_normal, buf, sizeof(buf));
                 if (flag_normal) {
                     lbxfont_select(2, 1, 0, 0);
                 }
@@ -208,21 +217,20 @@ static void planets_draw_cb(void *vptr)
                     lbxfont_print_str_normal(268, y0, str, UI_SCREEN_W, ui_scale);
                 }
             }
+            if (ui_extra_enabled) {
+                lbxfont_select(2, 6, 0, 0);
+            }
             if (p->slider[PLANET_SLIDER_SHIP] > 0) {
-                if (p->buildship == BUILDSHIP_STARGATE) {
-                    str = game_str_sm_stargate;
-                } else {
-                    str = g->srd[d->api].design[p->buildship].name;
-                }
+                str = planets_get_dock_str(g, p);
                 lbxfont_print_str_normal(221, y0, str, UI_SCREEN_W, ui_scale);
             }
         }
     }
 
     lbxfont_select(2, 6, 0, 0);
-    game_print_prod_of_total(g, d->api, e->ship_maint_bc, buf);
+    game_print_prod_of_total(g, d->api, e->ship_maint_bc, buf, sizeof(buf));
     lbxfont_print_str_right(59, 174, buf, UI_SCREEN_W, ui_scale);
-    game_print_prod_of_total(g, d->api, e->bases_maint_bc, buf);
+    game_print_prod_of_total(g, d->api, e->bases_maint_bc, buf, sizeof(buf));
     lbxfont_print_str_right(59, 185, buf, UI_SCREEN_W, ui_scale);
 
     v = 0;
@@ -231,14 +239,14 @@ static void planets_draw_cb(void *vptr)
             v += e->spying[i];
         }
     }
-    sprintf(buf, "%i.%i%%", v / 10, v % 10);
+    lib_sprintf(buf, sizeof(buf), "%i.%i%%", v / 10, v % 10);
     lbxfont_print_str_right(116, 174, buf, UI_SCREEN_W, ui_scale);
     v = e->security;
-    sprintf(buf, "%i.%i%%", v / 10, v % 10);
+    lib_sprintf(buf, sizeof(buf), "%i.%i%%", v / 10, v % 10);
     lbxfont_print_str_right(116, 185, buf, UI_SCREEN_W, ui_scale);
-    sprintf(buf, "%i %s", e->total_trade_bc, game_str_bc);
+    lib_sprintf(buf, sizeof(buf), "%i %s", e->total_trade_bc, game_str_bc);
     lbxfont_print_str_right(195, 174, buf, UI_SCREEN_W, ui_scale);
-    sprintf(buf, "%i %s", e->total_production_bc, game_str_bc);
+    lib_sprintf(buf, sizeof(buf), "%i %s", e->total_production_bc, game_str_bc);
     lbxfont_print_str_right(195, 185, buf, UI_SCREEN_W, ui_scale);
 
     lbxfont_select_set_12_1(5, 8, 0, 0);
@@ -262,7 +270,7 @@ static void planets_transfer_draw_cb(void *vptr)
     if (d->amount_trans > 0) {
         ui_draw_slider(x + 14, y + 36, d->amount_trans, 2, -1, 0x74, ui_scale);
     }
-    sprintf(buf, "%s %s", game_str_pl_resto, p->name);
+    lib_sprintf(buf, sizeof(buf), "%s %s", game_str_pl_resto, p->name);
     lbxfont_select(0, 0xd, 0, 0);
     lbxfont_print_str_center(x + 57, y + 11, game_str_pl_transof, UI_SCREEN_W, ui_scale);
     lbxfont_print_str_center(x + 57, y + 20, buf, UI_SCREEN_W, ui_scale);
@@ -279,7 +287,7 @@ static void ui_planets_transfer(struct planets_data_s *d)
 {
     struct game_s *g = d->g;
     planet_t *p = &(g->planet[d->planet_i]);
-    int16_t oi_cancel, oi_accept, oi_minus, oi_plus, oi_equals;
+    int16_t oi_cancel, oi_accept, oi_minus, oi_plus, oi_equals, oi_hash;
     int prod, allreserve, v;
     const int x = 100, y = 50;
     bool flag_done = false;
@@ -312,6 +320,7 @@ static void ui_planets_transfer(struct planets_data_s *d)
     oi_minus = uiobj_add_mousearea(x + 10, y + 33, x + 12, y + 41, MOO_KEY_UNKNOWN);
     oi_plus = uiobj_add_mousearea(x + 66, y + 33, x + 70, y + 41, MOO_KEY_UNKNOWN);
     oi_equals = uiobj_add_inputkey(MOO_KEY_EQUALS);
+    oi_hash = uiobj_add_inputkey(MOO_KEY_HASH);
 
     uiobj_set_callback_and_delay(planets_transfer_draw_cb, d, 1);
 
@@ -338,7 +347,7 @@ static void ui_planets_transfer(struct planets_data_s *d)
             ui_sound_play_sfx_24();
             d->amount_trans += 2;
             SETMIN(d->amount_trans, 100);
-        } else if (oi == oi_equals) {
+        } else if (oi == oi_equals || oi == oi_hash) {
             ui_sound_play_sfx_24();
             d->amount_trans = v;
             SETRANGE(d->amount_trans, 0, 100);
@@ -356,7 +365,6 @@ static void ui_planets_transfer(struct planets_data_s *d)
 enum {
     UI_SORT_INDEX = 0,
     UI_SORT_NAME,
-    UI_SORT_GOVERN,
     UI_SORT_POP,
     UI_SORT_GROWTH,
     UI_SORT_FACT,
@@ -378,7 +386,7 @@ enum {
     const planet_t *p0 = &(g->planet[pli0]); \
     const planet_t *p1 = &(g->planet[pli1])
 
-#define UI_SORT_CMP_VALUE(_v0_, _v1_) (((_v0_) != (_v1_)) ? ((_v0_) - (_v1_)) : (i0 - i1))
+#define UI_SORT_CMP_VALUE(_v0_, _v1_) (((_v0_) != (_v1_)) ? ((_v0_) - (_v1_)) : (i1 - i0))
 #define UI_SORT_CMP_VARIABLE(_var_) UI_SORT_CMP_VALUE(p0->_var_, p1->_var_)
 
 static int planets_sort_inc_index(const void *ptr0, const void *ptr1)
@@ -404,19 +412,6 @@ static int planets_sort_dec_name(const void *ptr0, const void *ptr1)
     return planets_sort_inc_name(ptr1, ptr0);
 }
 
-static int planets_sort_inc_govern(const void *ptr0, const void *ptr1)
-{
-    UI_SORT_SETUP();
-    uint8_t v0 = ui_draw_govern_color(p0, p0->owner);
-    uint8_t v1 = ui_draw_govern_color(p1, p1->owner);
-    return UI_SORT_CMP_VALUE(v0, v1);
-}
-
-static int planets_sort_dec_govern(const void *ptr0, const void *ptr1)
-{
-    return planets_sort_inc_govern(ptr1, ptr0);
-}
-
 static int planets_sort_inc_pop(const void *ptr0, const void *ptr1)
 {
     UI_SORT_SETUP();
@@ -455,7 +450,9 @@ static int planets_sort_dec_fact(const void *ptr0, const void *ptr1)
 static int planets_sort_inc_shield(const void *ptr0, const void *ptr1)
 {
     UI_SORT_SETUP();
-    return UI_SORT_CMP_VARIABLE(shield);
+    int v0 = p0->battlebg ? p0->shield : -1;
+    int v1 = p1->battlebg ? p1->shield : -1;
+    return UI_SORT_CMP_VALUE(v0, v1);
 }
 
 static int planets_sort_dec_shield(const void *ptr0, const void *ptr1)
@@ -501,9 +498,9 @@ static int planets_sort_dec_prod(const void *ptr0, const void *ptr1)
 static int planets_sort_inc_dock(const void *ptr0, const void *ptr1)
 {
     UI_SORT_SETUP();
-    int v0 = (p0->slider[PLANET_SLIDER_SHIP] > 0) ? p0->buildship : -1;
-    int v1 = (p1->slider[PLANET_SLIDER_SHIP] > 0) ? p1->buildship : -1;
-    return UI_SORT_CMP_VALUE(v0, v1);
+    const char *s0 = (p0->slider[PLANET_SLIDER_SHIP] > 0) ? planets_get_dock_str(g, p0) : "\xff";
+    const char *s1 = (p1->slider[PLANET_SLIDER_SHIP] > 0) ? planets_get_dock_str(g, p1) : "\xff";
+    return UI_SORT_CMP_VALUE(strcmp(s1, s0), 0);
 }
 
 static int planets_sort_dec_dock(const void *ptr0, const void *ptr1)
@@ -513,26 +510,26 @@ static int planets_sort_dec_dock(const void *ptr0, const void *ptr1)
 
 static int planets_sort_inc_notes(const void *ptr0, const void *ptr1)
 {
-    const struct game_s *g = ui_data.sorted.g;
-    uint16_t i0 = *((uint16_t const *)ptr0);
-    uint16_t i1 = *((uint16_t const *)ptr1);
-    uint8_t pli0 = ui_data.sorted.value[i0];
-    uint8_t pli1 = ui_data.sorted.value[i1];
+    UI_SORT_SETUP();
     char buf0[64];
     char buf1[64];
-    const char *s0 = planets_get_notes_str(g, pli0, 0, buf0);
-    const char *s1 = planets_get_notes_str(g, pli1, 0, buf1);
+    bool normal0, normal1;
+    const char *s0 = planets_get_notes_str(g, pli0, &normal0, buf0, sizeof(buf0));
+    const char *s1 = planets_get_notes_str(g, pli1, &normal1, buf1, sizeof(buf1));
     int d;
-    if ((!s0) && (!s1)) {
-        d = i0 - i1;
-    } else if ((!s0) && s1) {
-        d = -1;
-    } else if ((!s1) && s0) {
+    if (normal1 && !normal0) {
         d = 1;
+    } else if (normal0 && !normal1) {
+        d = -1;
+    } else if (!normal0 && !normal1) {
+        d = UI_SORT_CMP_VALUE(strcmp(s1, s0), 0);
     } else {
-        d = strcmp(planets_get_notes_str(g, pli0, 0, buf0), planets_get_notes_str(g, pli1, 0, buf1));
-        if (d == 0) {
-            d = i0 - i1;
+        if (p0->special > p1->special) {
+            d = 1;
+        } else if (p0->special < p1->special) {
+            d = -1;
+        } else {
+            d = UI_SORT_CMP_VARIABLE(growth);
         }
     }
     return d;
@@ -550,8 +547,6 @@ static sort_cb_t * const sort_cb_tbl[UI_SORT_NUM * 2] = {
     planets_sort_dec_index,
     planets_sort_inc_name,
     planets_sort_dec_name,
-    planets_sort_inc_govern,
-    planets_sort_dec_govern,
     planets_sort_dec_pop,
     planets_sort_inc_pop,
     planets_sort_dec_growth,
@@ -572,6 +567,23 @@ static sort_cb_t * const sort_cb_tbl[UI_SORT_NUM * 2] = {
     planets_sort_inc_notes
 };
 
+static void ui_planets_sort(struct planets_data_s *d)
+{
+    qsort(ui_data.sorted.index, d->num, sizeof(ui_data.sorted.index[0]), sort_cb_tbl[d->order_i]);
+    ui_data.sorted.planets_order[d->api] = d->order_i;
+    for (int i = 0; i < d->num; ++i) {
+        int index = ui_data.sorted.index[i];
+        if (ui_data.sorted.value[index] == d->focus_i) {
+            d->pos = i - 5;
+            break;
+        }
+    }
+    if ((d->pos + PLANETS_ON_SCREEN) >= d->num) {
+        d->pos = d->num - PLANETS_ON_SCREEN;
+    }
+    SETMAX(d->pos, 0);
+}
+
 /* -------------------------------------------------------------------------- */
 
 void ui_planets(struct game_s *g, player_id_t active_player)
@@ -604,6 +616,11 @@ void ui_planets(struct game_s *g, player_id_t active_player)
         }
     }
 
+    if (ui_extra_enabled) {
+        d.order_i = ui_data.sorted.planets_order[active_player];
+        ui_planets_sort(&d);
+    }
+
 again:
     flag_trans = false;
     game_update_production(g); /* this is needed for correct shown production after transfers */
@@ -686,7 +703,7 @@ again:
                     ++v;
                 }
                 d.order_i = v;
-                qsort(ui_data.sorted.index, d.num, sizeof(ui_data.sorted.index[0]), sort_cb_tbl[d.order_i]);
+                ui_planets_sort(&d);
             }
         }
         ui_cursor_setup_area(1, &ui_cursor_area_tbl[flag_trans ? 9 : 0]);
@@ -706,8 +723,8 @@ again:
                 }
             }
             if (ui_extra_enabled) {
-                const int x0[UI_SORT_NUM] = {  8, 23, 61, 71,  87, 117, 138, 155, 176, 194, 220, 267 };
-                const int x1[UI_SORT_NUM] = { 20, 60, 67, 86, 113, 134, 151, 172, 192, 216, 262, 310 };
+                const int x0[UI_SORT_NUM] = {  8, 23, 71,  87, 117, 138, 155, 176, 194, 220, 267 };
+                const int x1[UI_SORT_NUM] = { 20, 67, 86, 113, 134, 151, 172, 192, 216, 262, 310 };
                 const int y0 = 8, y1 = 16;
                 for (int i = 0; i < UI_SORT_NUM; ++i) {
                     oi_sort[i] = uiobj_add_mousearea(x0[i], y0, x1[i], y1, MOO_KEY_UNKNOWN);
