@@ -12,11 +12,14 @@
 #include "game_str.h"
 #include "game_tech.h"
 #include "kbd.h"
+#include "mouse.h"
 #include "lbxgfx.h"
 #include "lbxfont.h"
+#include "lib.h"
 #include "log.h"
 #include "rnd.h"
 #include "types.h"
+#include "ui.h"
 #include "uicursor.h"
 #include "uidraw.h"
 #include "uidefs.h"
@@ -32,14 +35,32 @@ const uint8_t colortbl_line_red[5] = { 0x44, 0x43, 0x42, 0x41, 0x40 };
 const uint8_t colortbl_line_reloc[5] = { 0x14, 0x15, 0x16, 0x17, 0x18 };
 const uint8_t colortbl_line_green[5] = { 0xb0, 0xb1, 0xb2, 0xb3, 0xb4 };
 
+static uint8_t ui_starmap_cursor_on_star(const struct starmap_data_s *d, int16_t oi)
+{
+    if (oi == 0) {
+        return PLANET_NONE;
+    }
+    for (int i = 0; i < d->g->galaxy_stars; ++i) {
+        if (oi == d->oi_tbl_stars[i]) {
+            return i;
+        }
+    }
+    return PLANET_NONE;
+}
+
 /* -------------------------------------------------------------------------- */
 
 static void ui_starmap_draw_planetinfo_do(const struct game_s *g, player_id_t api, uint8_t planet_i, bool explored, bool show_plus)
 {
     const planet_t *p = &g->planet[planet_i];
-    if (explored) {
-        lbxfont_select(4, 0xf, 0, 0);
+    if (explored || (ui_extra_enabled && g->gaux->flag_cheat_stars)) {
+        lbxfont_set_temp_color(0x0);
+        lbxfont_select_set_12_4(4, 0xf, 0, 0);
         lbxfont_print_str_center(269, 10, p->name, UI_SCREEN_W, ui_scale);
+        /* stars in nebulas get a purple instead of a red frame */
+        if (ui_extra_enabled && p->battlebg==0) {
+            ui_draw_box1(228, 25, 309, 52, 0xd3, 0xd3, ui_scale);
+        }
         if (p->type == PLANET_TYPE_NOT_HABITABLE) {
             lbxfont_select(0, 0xe, 0, 0);
             lbxfont_print_str_center(269, 32, game_str_sm_nohabit, UI_SCREEN_W, ui_scale);
@@ -77,7 +98,7 @@ static void ui_starmap_draw_planetinfo_do(const struct game_s *g, player_id_t ap
                 }
                 lbxfont_print_str_right(305, 36, str, UI_SCREEN_W, ui_scale);
                 lbxfont_select(2, 0xe, 0, 0);
-                if (show_plus && (p->owner == api) && ((p->max_pop2 + g->eto[api].have_terraform_n) > max_pop) && (max_pop < game_num_max_pop)) {
+                if (show_plus && ((p->owner == api) || (ui_extra_enabled && g->gaux->flag_cheat_stars)) && game_planet_can_terraform(g, p, p->owner, ui_extra_enabled)) {
                     lbxfont_print_str_normal(289, 44, "+", UI_SCREEN_W, ui_scale);
                     x = 287;
                     xp = x - 22;
@@ -105,7 +126,7 @@ static void ui_starmap_draw_range_parsec(struct starmap_data_s *d, int y)
     const struct game_s *g = d->g;
     int dist = game_get_min_dist(g, d->api, g->planet_focus_i[d->api]);
     char buf[64];
-    sprintf(buf, "%s %i %s", game_str_sm_range, dist, (dist == 1) ? game_str_sm_parsec : game_str_sm_parsecs);
+    lib_sprintf(buf, sizeof(buf), "%s %i %s", game_str_sm_range, dist, (dist == 1) ? game_str_sm_parsec : game_str_sm_parsecs);
     lbxfont_select_set_12_4(0, 4, 0, 0);
     lbxfont_print_str_center(269, y, buf, UI_SCREEN_W, ui_scale);
 }
@@ -122,15 +143,8 @@ static void ui_starmap_draw_sliders_and_prod(struct starmap_data_s *d)
     }
 
     lbxgfx_draw_frame(224, 5, ui_data.gfx.starmap.yourplnt, UI_SCREEN_W, ui_scale);
-    {
-        /* highlight governor with a border; use a color based on mode */
-        uint8_t c = ui_draw_govern_color(p, d->api);
-        if (c != 0) {
-            ui_draw_box1(227, 8, 310, 20, c, c, ui_scale);
-        }
-    }
     lbxfont_select(2, 0xd, 0xe, 0);
-    sprintf(buf, "%i \x02(%i)\x01", p->prod_after_maint, p->total_prod);
+    lib_sprintf(buf, sizeof(buf), "%i \x02(%i)\x01", p->prod_after_maint, p->total_prod);
     lbxfont_print_str_right(x, 72, buf, UI_SCREEN_W, ui_scale);
     lbxfont_select(2, 0xd, 0, 0);
     lbxfont_print_num_right(265, 61, p->pop, UI_SCREEN_W, ui_scale);
@@ -149,7 +163,7 @@ static void ui_starmap_draw_sliders_and_prod(struct starmap_data_s *d)
         lbxgfx_draw_frame(229, 141, ui_data.gfx.starmap.stargate, UI_SCREEN_W, ui_scale);
         lbxfont_print_str_center(251, 169, game_str_sm_stargate, UI_SCREEN_W, ui_scale);
     } else {
-        const shipdesign_t *sd = &g->srd[d->api].design[p->buildship];
+        const shipdesign_t *sd = &g->srd[p->owner].design[p->buildship];
         uint8_t *gfx = ui_data.gfx.ships[sd->look];
         lbxgfx_set_frame_0(gfx);
         lbxgfx_draw_frame(236, 142, gfx, UI_SCREEN_W, ui_scale);
@@ -159,7 +173,7 @@ static void ui_starmap_draw_sliders_and_prod(struct starmap_data_s *d)
     lbxfont_select(2, 6, 0, 0);
     {
         int v;
-        v = game_planet_get_slider_text(g, p, d->api, PLANET_SLIDER_SHIP, buf);
+        v = game_planet_get_slider_text(g, p, PLANET_SLIDER_SHIP, buf, sizeof(buf));
         lbxfont_print_str_right(x, 83, buf, UI_SCREEN_W, ui_scale);
         lbxfont_select(0, 0xd, 0, 0);
         if (v >= 0) {
@@ -167,30 +181,30 @@ static void ui_starmap_draw_sliders_and_prod(struct starmap_data_s *d)
         }
     }
     lbxfont_select(2, 6, 0, 0);
-    game_planet_get_slider_text(g, p, d->api, PLANET_SLIDER_DEF, buf);
+    game_planet_get_slider_text(g, p, PLANET_SLIDER_DEF, buf, sizeof(buf));
     lbxfont_print_str_right(x, 94, buf, UI_SCREEN_W, ui_scale);
-    game_planet_get_slider_text(g, p, d->api, PLANET_SLIDER_IND, buf);
+    game_planet_get_slider_text(g, p, PLANET_SLIDER_IND, buf, sizeof(buf));
     lbxfont_print_str_right(x, 105, buf, UI_SCREEN_W, ui_scale);
     {
         int v;
-        v = game_planet_get_slider_text_eco(g, p, d->api, ui_extra_enabled, buf);
+        v = game_planet_get_slider_text_eco(g, p, ui_extra_enabled, buf, sizeof(buf));
         lbxfont_print_str_right(x, 116, buf, UI_SCREEN_W, ui_scale);
         if (v >= 0) {
             if (ui_extra_enabled) {
                 if (v < 100) {
-                    sprintf(buf, "%i.%i", v / 10, v % 10); /* "+0.X" does not fit the box */
+                    lib_sprintf(buf, sizeof(buf), "%i.%i", v / 10, v % 10); /* "+0.X" does not fit the box */
                 } else {
-                    sprintf(buf, "+%i", v / 10);
+                    lib_sprintf(buf, sizeof(buf), "+%i", v / 10);
                 }
             } else {
-                sprintf(buf, "+%i", v);
+                lib_sprintf(buf, sizeof(buf), "+%i", v);
             }
             lbxfont_print_str_right(297, 116, buf, UI_SCREEN_W, ui_scale);
         }
     }
     {
         int v;
-        v = game_planet_get_slider_text(g, p, d->api, PLANET_SLIDER_TECH, buf);
+        v = game_planet_get_slider_text(g, p, PLANET_SLIDER_TECH, buf, sizeof(buf));
         if (v > 9999) {
             ui_draw_filled_rect(288, 127, 312, 132, 7, ui_scale);
         } else {
@@ -210,8 +224,9 @@ static void ui_starmap_draw_textbox_finished(const struct game_s *g, player_id_t
             break;
         }
     }
-    game_planet_get_finished_text(g, p, i, buf);
-    ui_draw_textbox_2str("", buf, 54, ui_scale);
+    game_planet_get_finished_text(g, p, i, buf, UI_STRBUF_SIZE);
+    int y = ui_extra_enabled ? 45 : 54;
+    ui_draw_textbox_2str("", buf, y, ui_scale);
     ui_draw_textbox_2str("", game_str_sm_planratio, 110, ui_scale);
 }
 
@@ -244,17 +259,9 @@ static int ui_starmap_scrollkey_accel(int zh)
     if (zh < 0) {
         v = -v;
     }
-    if (v > 64) {
-        v = 5;
-    } if (v > 32) {
-        v = 4;
-    } else if (v > 16) {
-        v = 3;
-    } else if (v > 4) {
-        v = 2;
-    } else {
-        v = 1;
-    }
+    v = 1 + (v / 4);
+    if (v > 8) v = 8;
+    v *= ui_sm_scroll_speed / 2 + 1;
     if (zh < 0) {
         v = -v;
     }
@@ -271,7 +278,7 @@ void ui_starmap_draw_basic(struct starmap_data_s *d)
     ui_starmap_draw_starmap(d);
     ui_starmap_draw_button_text(d, true);
     ui_draw_filled_rect(224, 5, 314, 178, 0, ui_scale);
-    if (BOOLVEC_IS0(p->explored, d->api)) {
+    if (BOOLVEC_IS0(p->explored, d->api) && !(ui_extra_enabled && g->gaux->flag_cheat_stars)) {
         lbxgfx_draw_frame(224, 5, ui_data.gfx.starmap.unexplor, UI_SCREEN_W, ui_scale);
         lbxfont_select_set_12_4(5, 1, 0, 0);
         lbxfont_print_str_split(232, 74, 76, game_str_tbl_sm_stinfo[p->star_type], 2, UI_SCREEN_W, UI_SCREEN_H, ui_scale);
@@ -279,7 +286,7 @@ void ui_starmap_draw_basic(struct starmap_data_s *d)
     } else {
         player_id_t owner = p->owner;
         int pi = g->planet_focus_i[d->api];
-        if (BOOLVEC_IS0(p->within_srange, d->api) && ((owner == PLAYER_NONE) || BOOLVEC_IS0(g->eto[d->api].contact, owner))) {
+        if (BOOLVEC_IS0(p->within_srange, d->api) && !(ui_extra_enabled && g->gaux->flag_cheat_stars) && ((owner == PLAYER_NONE) || BOOLVEC_IS0(g->eto[d->api].contact, owner))) {
             owner = g->seen[d->api][pi].owner;
         }
         if (owner == PLAYER_NONE) {
@@ -288,14 +295,14 @@ void ui_starmap_draw_basic(struct starmap_data_s *d)
             lbxgfx_draw_frame(227, 73, ui_data.gfx.colonies.current, UI_SCREEN_W, ui_scale);
             ui_draw_box1(227, 73, 310, 174, 0, 0, ui_scale);
             ui_starmap_draw_range_parsec(d, 80);
-        } else if ((owner != d->api) || (p->unrest == PLANET_UNREST_REBELLION)) {
+        } else if (((owner != d->api) && !(ui_extra_enabled && g->gaux->flag_cheat_stars)) || (p->unrest == PLANET_UNREST_REBELLION)) {
             char buf[64];
             int pop, bases, range_y;
             lbxgfx_draw_frame(224, 5, ui_data.gfx.starmap.en_colny, UI_SCREEN_W, ui_scale);
             ui_data.gfx.colonies.current = ui_data.gfx.colonies.d[p->type * 2 + 1];
             lbxgfx_draw_frame(227, 73, ui_data.gfx.colonies.current, UI_SCREEN_W, ui_scale);
             ui_draw_box1(227, 73, 310, 174, 0, 0, ui_scale);
-            sprintf(buf, "%s %s", game_str_tbl_race[g->eto[owner].race], game_str_sm_colony);
+            lib_sprintf(buf, sizeof(buf), "%s %s", game_str_tbl_race[g->eto[owner].race], game_str_sm_colony);
             if (BOOLVEC_IS1(p->within_srange, d->api)) {
                 lbxfont_select_set_12_4(5, tbl_banner_fontparam[g->eto[owner].banner], 0, 0);
                 lbxfont_print_str_center(270, 84, buf, UI_SCREEN_W, ui_scale);
@@ -341,6 +348,7 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
 {
     const struct game_s *g = d->g;
     int x, y, tx, ty;
+    char str[16];
     STARMAP_LIM_INIT();
     {
         int v, step;
@@ -381,7 +389,7 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
     if (++d->anim_delay >= STARMAP_ANIM_DELAY) {
         d->anim_delay = 0;
     }
-    ui_draw_filled_rect(STARMAP_LIMITS, 0, starmap_scale);
+    ui_draw_filled_rect(6, 6, 222 - 1, 178 - 1, 0, ui_scale);
     lbxgfx_draw_frame(0, 0, ui_data.gfx.starmap.mainview, UI_SCREEN_W, ui_scale);
     uiobj_set_limits(STARMAP_LIMITS);
     {
@@ -394,12 +402,20 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
         ly = 178 * ui_scale - 1;
         for (int yb = y0; yb < ly; yb += 200) {
             for (int xb = x0; xb < lx; xb += 320) {
-                lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starback, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                if (ui_fixbugs_enabled) {
+                    lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starbak2, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                } else {
+                    lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starback, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                }
             }
         }
         for (int yb = y1; yb < ly; yb += 200) {
             for (int xb = x1; xb < lx; xb += 320) {
-                lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starbak2, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                if (ui_fixbugs_enabled) {
+                    lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starback, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                } else {
+                    lbxgfx_draw_frame_offs(xb, yb, ui_data.gfx.starmap.starbak2, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+                }
             }
         }
     }
@@ -411,14 +427,14 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
     }
     if (ui_data.starmap.flag_show_grid) {
         int x0, y0, x1, y1;
-        for (y0 = 10; y0 < (g->galaxy_maxy * 2); y0 += 50) {
+        for (y0 = 10; y0 < g->galaxy_maxy; y0 += 50) {
             int ty;
             x0 = (-x) * 2 + 6;
             x1 = (g->galaxy_maxx - x) * 2 + 6;
             ty = (y0 - y) * 2 + 6;
             ui_draw_line_limit(x0, ty, x1, ty, 4, starmap_scale);
         }
-        for (x0 = 10; x0 < (g->galaxy_maxx * 2); x0 += 50) {
+        for (x0 = 10; x0 < g->galaxy_maxx; x0 += 50) {
             int tx;
             y0 = (-y) * 2 + 6;
             y1 = (g->galaxy_maxy - y) * 2 + 6;
@@ -442,6 +458,9 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
         const planet_t *p = &g->planet[pi];
         uint8_t *gfx = ui_data.gfx.starmap.stars[p->star_type + p->look];
         uint8_t anim_frame = ui_data.star_frame[pi];
+        bool explored = BOOLVEC_IS1(p->explored, d->api);
+        bool visible = BOOLVEC_IS1(p->within_srange, d->api);
+        bool done = false;
         lbxgfx_set_new_frame(gfx, (anim_frame < 4) ? anim_frame : 0);
         gfx_aux_draw_frame_to(gfx, &ui_data.starmap.star_aux);
         if (p->look > 0) {
@@ -460,20 +479,54 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
             }
             ui_data.star_frame[pi] = anim_frame;
         }
-        if (p->owner != PLAYER_NONE) {
-            bool do_print;
-            do_print = BOOLVEC_IS1(p->within_srange, d->api);
-            if (do_print || (pi == g->evn.planet_orion_i)) {
-                do_print = true;
-                lbxfont_select(2, tbl_banner_fontparam[g->eto[p->owner].banner], 0, 0);
-            } else if (BOOLVEC_IS1(g->eto[d->api].contact, p->owner) || (p->within_frange[d->api] == 1)) {
-                do_print = true;
-                lbxfont_select(2, 0, 0, 0);
-                lbxfont_set_color0(tbl_banner_color[g->eto[p->owner].banner]);
+        tx = (p->x - x) * 2 + 14;
+        ty = (p->y - y) * 2 + 22;
+        if (p->owner == PLAYER_NONE) {
+            lbxfont_select(2, 7, 0, 0);
+        } else if (visible || (pi == g->evn.planet_orion_i)) {
+            lbxfont_select(2, tbl_banner_fontparam[g->eto[p->owner].banner], 0, 0);
+        } else if (BOOLVEC_IS1(g->eto[d->api].contact, p->owner) || (p->within_frange[d->api] == 1)) {
+            lbxfont_select(2, 0, 0, 0);
+            lbxfont_set_color0(tbl_banner_color[g->eto[p->owner].banner]);
+        } else {
+            lbxfont_select(2, 7, 0, 0);
+        }
+        if (ui_extra_enabled && explored && ui_data.starmap.star_text_type != UI_SM_STAR_TEXT_NAME) {
+            if (p->type == PLANET_TYPE_NOT_HABITABLE) {
+                lbxfont_print_str_center_limit(tx, ty, "0/0", STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
+                done = true;
+            } else if (visible && ui_data.starmap.star_text_type == UI_SM_STAR_TEXT_POPULATION) {
+                int max_pop = p->max_pop3;
+                if (g->eto[d->api].race != RACE_SILICOID) {
+                    max_pop -= p->waste;
+                }
+                if (game_planet_can_terraform(g, p, d->api, true)) {
+                    lib_sprintf(str, sizeof(str), "%i/%i+", p->pop, max_pop);
+                } else {
+                    lib_sprintf(str, sizeof(str), "%i/%i", p->pop, max_pop);
+                }
+                lbxfont_print_str_center_limit(tx, ty, str, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
+                done = true;
+            } else if (ui_data.starmap.star_text_type == UI_SM_STAR_TEXT_ENVIRONMENT) {
+                lib_sprintf(str, sizeof(str), "%s", game_str_tbl_sm_pltype[p->type]);
+                lbxfont_print_str_center_limit(tx, ty, str, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
+                done = true;
+            } else if (ui_data.starmap.star_text_type == UI_SM_STAR_TEXT_SPECIAL) {
+                if (p->special == PLANET_SPECIAL_NORMAL) {
+                    lib_sprintf(str, sizeof(str), "%s", "Normal");
+                } else {
+                    lib_sprintf(str, sizeof(str), "%s", game_str_tbl_sm_pspecial[p->special]);
+                }
+                lbxfont_print_str_center_limit(tx, ty, str, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
+                done = true;
             }
-            if (do_print) {
-                tx = (p->x - x) * 2 + 14;
-                ty = (p->y - y) * 2 + 22;
+        }
+        if (!done) {
+            bool do_print = visible
+                        || (pi == g->evn.planet_orion_i)
+                        || (BOOLVEC_IS1(g->eto[d->api].contact, p->owner))
+                        || (p->within_frange[d->api] == 1);
+            if (p->owner != PLAYER_NONE && do_print) {
                 lbxfont_print_str_center_limit(tx, ty, p->name, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
             }
         }
@@ -489,7 +542,7 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
         const planet_t *p = &g->planet[g->planet_focus_i[d->api]];
         tx = (p->x - x) * 2 + 8;
         ty = (p->y - y) * 2 + 8;
-        lbxgfx_draw_frame_offs(tx, ty, ui_data.gfx.starmap.planbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+        lbxgfx_draw_frame_offs_delay(tx, ty, !d->anim_delay, ui_data.gfx.starmap.planbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
     }
     if (d->anim_delay == 0) {
         if (--ui_data.starmap.line_anim_phase < 0) {
@@ -512,6 +565,11 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
             if (g->eto[d->api].have_ia_scanner && (p->owner == d->api) && (r->owner != d->api) && (r->dest == g->planet_focus_i[d->api])) {
                 ui_draw_line_limit_ctbl(tx + 5, ty + 2, (p->x - x) * 2 + 14, (p->y - y) * 2 + 14, colortbl_line_red, 5, ui_data.starmap.line_anim_phase, starmap_scale);
             }
+            if (ui_extra_enabled && ui_data.starmap.flag_show_own_routes && d->show_planet_focus) {
+                if ((r->owner == d->api) && (r->dest == g->planet_focus_i[d->api])) {
+                    ui_draw_line_limit_ctbl(tx + 5, ty + 2, (p->x - x) * 2 + 14, (p->y - y) * 2 + 14, colortbl_line_green, 5, ui_data.starmap.line_anim_phase, starmap_scale);
+                }
+            }
             lbxgfx_draw_frame_offs(tx, ty, gfx, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
         }
     }
@@ -531,6 +589,11 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
             if (g->eto[d->api].have_ia_scanner && (p->owner == d->api) && (r->owner != d->api) && (r->dest == g->planet_focus_i[d->api])) {
                 ui_draw_line_limit_ctbl(tx + 5, ty + 2, (p->x - x) * 2 + 14, (p->y - y) * 2 + 14, colortbl_line_red, 5, ui_data.starmap.line_anim_phase, starmap_scale);
             }
+            if (ui_extra_enabled && ui_data.starmap.flag_show_own_routes && d->show_planet_focus) {
+                if ((r->owner == d->api) && (r->dest == g->planet_focus_i[d->api])) {
+                    ui_draw_line_limit_ctbl(tx + 5, ty + 2, (p->x - x) * 2 + 14, (p->y - y) * 2 + 14, colortbl_line_green, 5, ui_data.starmap.line_anim_phase, starmap_scale);
+                }
+            }
             lbxgfx_draw_frame_offs(tx, ty, gfx, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
         }
     }
@@ -539,14 +602,14 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
         ty = (g->evn.crystal.y - y) * 2 + 8;
         lbxgfx_draw_frame_offs(tx, ty, ui_data.gfx.planets.smonster, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
         lbxfont_select(2, 8, 0, 0);
-        lbxfont_print_str_center_limit(tx + 2, ty + 5, game_str_sm_crystal, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+        lbxfont_print_str_center_limit(tx + 2, ty + 5, game_str_sm_crystal, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
     }
     if ((g->evn.amoeba.exists != 0) && (g->evn.amoeba.killer == PLAYER_NONE)) {
         tx = (g->evn.amoeba.x - x) * 2 + 8;
         ty = (g->evn.amoeba.y - y) * 2 + 8;
         lbxgfx_draw_frame_offs(tx, ty, ui_data.gfx.planets.smonster, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
         lbxfont_select(2, 8, 0, 0);
-        lbxfont_print_str_center_limit(tx + 2, ty + 5, game_str_sm_amoeba, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+        lbxfont_print_str_center_limit(tx + 2, ty + 5, game_str_sm_amoeba, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
     }
     for (int pi = 0; pi < g->galaxy_stars; ++pi) {
         const planet_t *p = &g->planet[pi];
@@ -580,6 +643,16 @@ void ui_starmap_draw_starmap(struct starmap_data_s *d)
             }
         }
     }
+    if (ui_sm_distance_tooltip && ui_data.ui_main_loop_action == UI_MAIN_LOOP_STARMAP && d->dist_i != PLANET_NONE) {
+        int j = g->planet_focus_i[d->api];
+        const planet_t *p = &g->planet[d->dist_i], *q = &g->planet[j];
+        int px = 2 * (p->x - x) + 14, py = 2 * (p->y - y) + 14;
+        int qx = 2 * (q->x - x) + 14, qy = 2 * (q->y - y) + 14;
+        ui_draw_line_limit(px, py, qx, qy, 0x06, starmap_scale);
+        lbxfont_select(0, 0, 0, 0);
+        int l = g->gaux->star_dist[j][d->dist_i];
+        lbxfont_print_num_center_limit( (px + qx) / 2, (py + qy) / 2 - 2, l, STARMAP_TEXT_LIMITS, UI_SCREEN_W, starmap_scale);
+    }
 }
 
 void ui_starmap_draw_button_text(struct starmap_data_s *d, bool highlight)
@@ -602,6 +675,45 @@ void ui_starmap_draw_button_text(struct starmap_data_s *d, bool highlight)
     lbxfont_print_str_normal(263, 184, game_str_sm_next_turn, UI_SCREEN_W, ui_scale);
 }
 
+static void ui_starmap_clamp_xy(const struct game_s *g, int *x, int *y)
+{
+    if (!ui_sm_expanded_scroll) {
+        SETRANGE(*x, 0, g->galaxy_maxx - ((108 * ui_scale) / starmap_scale));
+        SETRANGE(*y, 0, g->galaxy_maxy - ((86 * ui_scale) / starmap_scale));
+    } else {
+        SETRANGE(*x, -((54 * ui_scale) / starmap_scale), g->galaxy_maxx - ((54 * ui_scale) / starmap_scale));
+        SETRANGE(*y, -((43 * ui_scale) / starmap_scale), g->galaxy_maxy - ((43 * ui_scale) / starmap_scale));
+    }
+}
+
+void ui_starmap_set_pos_focus(const struct game_s *g, player_id_t active_player)
+{
+    const planet_t *p = &g->planet[g->planet_focus_i[active_player]];
+    ui_starmap_set_pos(g, p->x, p->y);
+}
+
+void ui_starmap_set_pos(const struct game_s *g, int x, int y)
+{
+    x -= (54 * ui_scale) / starmap_scale;
+    y -= (43 * ui_scale) / starmap_scale;
+    ui_starmap_clamp_xy(g, &x, &y);
+    ui_data.starmap.x = x;
+    ui_data.starmap.x2 = x;
+    ui_data.starmap.y = y;
+    ui_data.starmap.y2 = y;
+}
+
+static void ui_starmap_select_target(struct starmap_data_s *d, uint8_t planet_i)
+{
+    if (!d->controllable || (planet_i == PLANET_NONE)) {
+        return;
+    }
+    d->g->planet_focus_i[d->api] = planet_i;
+    if (ui_data.ui_main_loop_action == UI_MAIN_LOOP_TRANS) {
+        d->tr.other = true;
+    }
+}
+
 void ui_starmap_handle_oi_ctrl(struct starmap_data_s *d, int16_t oi)
 {
 #define XSTEP   0x1b
@@ -614,7 +726,16 @@ void ui_starmap_handle_oi_ctrl(struct starmap_data_s *d, int16_t oi)
     }
     x = ui_data.starmap.x;
     y = ui_data.starmap.y;
-    if (oi == d->oi_ctrl_ul) {
+    if (oi == d->oi_scroll) {
+        if (d->scrollx >= 0) {
+            x += d->scrollx - 54;
+            y += d->scrolly - 43;
+            changed = true;
+        } else {
+            starmap_scale = d->scrollz;
+            d->set_pos_focus(g, g->active_player);
+        }
+    } else if (oi == d->oi_ctrl_ul) {
         x -= XSTEP;
         y -= YSTEP;
         changed = true;
@@ -642,10 +763,19 @@ void ui_starmap_handle_oi_ctrl(struct starmap_data_s *d, int16_t oi)
         x += XSTEP;
         y += YSTEP;
         changed = true;
+    } else if (oi == d->oi_pgdown) {
+        --d->scrollz;
+        SETMAX(d->scrollz, 1);
+        starmap_scale = d->scrollz;
+        ui_starmap_set_pos_focus(g, g->active_player);
+    } else if (oi == d->oi_pgup) {
+        ++d->scrollz;
+        SETMIN(d->scrollz, ui_scale);
+        starmap_scale = d->scrollz;
+        ui_starmap_set_pos_focus(g, g->active_player);
     }
     if (changed) {
-        SETRANGE(x, 0, g->galaxy_maxx - ((108 * ui_scale) / starmap_scale));
-        SETRANGE(y, 0, g->galaxy_maxy - ((86 * ui_scale) / starmap_scale));
+        ui_starmap_clamp_xy(g, &x, &y);
         ui_data.starmap.x2 = x;
         ui_data.starmap.y2 = y;
     }
@@ -669,12 +799,16 @@ void ui_starmap_handle_scrollkeys(struct starmap_data_s *d, int16_t oi)
     y = ui_data.starmap.y;
     xh = ui_data.starmap.xhold;
     yh = ui_data.starmap.yhold;
-    if (kbd_is_pressed(MOO_KEY_u, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)) {
+    if (0
+      || kbd_is_pressed(MOO_KEY_u, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)
+      || (ui_sm_mouse_scroll && (moouse_y <= 0))) {
         if (yh > 0) {
             yh = 0;
         }
         --yh;
-    } else if (kbd_is_pressed(MOO_KEY_j, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)) {
+    } else if (0
+      || kbd_is_pressed(MOO_KEY_j, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)
+      || (ui_sm_mouse_scroll && (moouse_y >= UI_SCREEN_H - 1))) {
         if (yh < 0) {
             yh = 0;
         }
@@ -685,12 +819,16 @@ void ui_starmap_handle_scrollkeys(struct starmap_data_s *d, int16_t oi)
     if (yh) {
         y += ui_starmap_scrollkey_accel(yh);
     }
-    if (kbd_is_pressed(MOO_KEY_h, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)) {
+    if (0
+      || kbd_is_pressed(MOO_KEY_h, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)
+      || (ui_sm_mouse_scroll && (moouse_x <= 0))) {
         if (xh > 0) {
             xh = 0;
         }
         --xh;
-    } else if (kbd_is_pressed(MOO_KEY_k, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)) {
+    } else if (0
+      || kbd_is_pressed(MOO_KEY_k, 0, MOO_MOD_SHIFT | MOO_MOD_ALT | MOO_MOD_CTRL)
+      || (ui_sm_mouse_scroll && (moouse_x >= UI_SCREEN_W - 1))) {
         if (xh < 0) {
             xh = 0;
         }
@@ -702,8 +840,7 @@ void ui_starmap_handle_scrollkeys(struct starmap_data_s *d, int16_t oi)
         x += ui_starmap_scrollkey_accel(xh);
     }
     if (xh || yh) {
-        SETRANGE(x, 0, g->galaxy_maxx - ((108 * ui_scale) / starmap_scale));
-        SETRANGE(y, 0, g->galaxy_maxy - ((86 * ui_scale) / starmap_scale));
+        ui_starmap_clamp_xy(g, &x, &y);
         ui_data.starmap.x2 = x;
         ui_data.starmap.y2 = y;
         ui_data.starmap.x = x;
@@ -713,41 +850,106 @@ void ui_starmap_handle_scrollkeys(struct starmap_data_s *d, int16_t oi)
     ui_data.starmap.yhold = yh;
 }
 
-uint8_t ui_starmap_handle_tag(struct starmap_data_s *d, int16_t oi, bool flag_set_focus)
-{
-    for (int i = 0; i < PLANET_TAG_NUM; ++i) {
-        if (oi == d->oi_tag_set[i]) {
-            ui_data.starmap.tag[d->api][i] = d->g->planet_focus_i[d->api];
-            break;
-        }
-        if (oi == d->oi_tag_get[i]) {
-            uint8_t pli;
-            pli = ui_data.starmap.tag[d->api][i];
-            if (pli < d->g->galaxy_stars) {
-                d->g->planet_focus_i[d->api] = pli;
-                if (flag_set_focus) {
-                    ui_starmap_set_pos_focus(d->g, d->api);
-                }
-                return pli;
-            }
-            break;
-        }
-    }
-    return PLANET_NONE;
-}
-
 void ui_starmap_add_oi_bottom_buttons(struct starmap_data_s *d)
 {
     d->oi_gameopts = uiobj_add_mousearea(5, 181, 36, 194, MOO_KEY_g);
     d->oi_design = uiobj_add_mousearea(40, 181, 75, 194, MOO_KEY_d);
     d->oi_fleet = uiobj_add_mousearea(79, 181, 111, 194, MOO_KEY_f);
     d->oi_map = uiobj_add_mousearea(115, 181, 139, 194, MOO_KEY_m);
-    d->oi_races = uiobj_add_mousearea(143, 181, 176, 194, MOO_KEY_r);
+    d->oi_races = uiobj_add_mousearea(143, 181, 176, 194, (ui_illogical_hotkey_fix ? MOO_KEY_a : MOO_KEY_r));
     d->oi_planets = uiobj_add_mousearea(180, 181, 221, 194, MOO_KEY_p);
     d->oi_tech = uiobj_add_mousearea(225, 181, 254, 194, MOO_KEY_t);
     d->oi_next_turn = uiobj_add_mousearea(258, 181, 314, 194, MOO_KEY_n);
 }
 
+bool ui_starmap_handle_oi_bottom_buttons(struct starmap_data_s *d, int16_t oi)
+{
+    ui_main_loop_action_t action = UI_MAIN_LOOP_NUM;
+    if (oi == d->oi_gameopts) {
+        action = UI_MAIN_LOOP_GAMEOPTS;
+    } else if (oi == d->oi_design) {
+        action = UI_MAIN_LOOP_DESIGN;
+    } else if (oi == d->oi_fleet) {
+        action = UI_MAIN_LOOP_FLEET;
+    } else if (oi == d->oi_map) {
+        action = UI_MAIN_LOOP_MAP;
+    } else if (oi == d->oi_races) {
+        action = UI_MAIN_LOOP_RACES;
+    } else if (oi == d->oi_planets) {
+        action = UI_MAIN_LOOP_PLANETS;
+    } else if (oi == d->oi_tech) {
+        action = UI_MAIN_LOOP_TECH;
+    } else if (oi == d->oi_next_turn) {
+        action = UI_MAIN_LOOP_NEXT_TURN;
+    }
+    if (action != UI_MAIN_LOOP_NUM) {
+        ui_data.ui_main_loop_action = action;
+        return true;
+    }
+    return false;
+}
+
+void ui_starmap_add_oi_misc(struct starmap_data_s *d)
+{
+    d->oi_alt_c = uiobj_add_inputkey(MOO_KEY_c | MOO_MOD_ALT);
+    d->oi_alt_m = uiobj_add_inputkey(MOO_KEY_m | MOO_MOD_ALT);
+    if (d->show_planet_focus) {
+        d->oi_alt_r = uiobj_add_inputkey(MOO_KEY_r | MOO_MOD_ALT);
+        d->oi_ctrl_r = uiobj_add_inputkey(MOO_KEY_r | MOO_MOD_CTRL);
+    }
+    if (ui_extra_enabled) {
+        d->oi_alt_f = uiobj_add_inputkey(MOO_KEY_f | MOO_MOD_ALT);
+        d->oi_alt_o = uiobj_add_inputkey(MOO_KEY_o | MOO_MOD_ALT);
+    }
+}
+
+bool ui_starmap_handle_oi_misc(struct starmap_data_s *d, int16_t oi)
+{
+    bool match = false;
+    struct game_s *g = d->g;
+    uint8_t planet_focus_i = g->planet_focus_i[d->api];
+    if (oi == d->oi_alt_m) {
+        ui_data.starmap.flag_show_grid = !ui_data.starmap.flag_show_grid;
+        match = true;
+    } else if (oi == d->oi_alt_c) {
+        d->set_pos_focus(d->g, d->api);
+        match = true;
+    } else if (oi == d->oi_alt_f) {
+        ui_data.starmap.flag_show_own_routes = !ui_data.starmap.flag_show_own_routes;
+        match = true;
+    } else if (oi == d->oi_alt_o) {
+        ui_data.starmap.star_text_type = (ui_data.starmap.star_text_type + 1) % UI_SM_STAR_TEXT_NUM;
+        match = true;
+    } else if ((oi == d->oi_alt_r) && game_reloc_dest_ok(g, planet_focus_i, d->api)) {
+        for (int i = 0; i < g->galaxy_stars; ++i) {
+            planet_t *p = &(g->planet[i]);
+            if ((p->owner == d->api) && (p->reloc != i)) {
+                p->reloc = planet_focus_i;
+            }
+        }
+        match = true;
+    } else if ((oi == d->oi_ctrl_r) && game_reloc_dest_ok(g, planet_focus_i, d->api)) {
+        int count = 0;
+        for (int i = 0; i < g->galaxy_stars; ++i) {
+            planet_t *p = &(g->planet[i]);
+            if ((p->owner == d->api) && (p->reloc != planet_focus_i)) {
+                p->reloc = planet_focus_i;
+                ++count;
+            }
+        }
+        if (count == 0) {
+            for (int i = 0; i < g->galaxy_stars; ++i) {
+                planet_t *p = &(g->planet[i]);
+                if (p->owner == d->api) {
+                    p->reloc = i;
+                }
+            }
+        }
+        match = true;
+    }
+    return match;
+}
+
 void ui_starmap_fill_oi_tbls(struct starmap_data_s *d)
 {
     const struct game_s *g = d->g;
@@ -827,6 +1029,7 @@ void ui_starmap_fill_oi_tbl_stars_own(struct starmap_data_s *d, player_id_t owne
 
 void ui_starmap_clear_oi_ctrl(struct starmap_data_s *d)
 {
+    d->oi_scroll = UIOBJI_INVALID;
     d->oi_ctrl_left = UIOBJI_INVALID;
     d->oi_ctrl_l2 = UIOBJI_INVALID;
     d->oi_ctrl_right = UIOBJI_INVALID;
@@ -839,12 +1042,13 @@ void ui_starmap_clear_oi_ctrl(struct starmap_data_s *d)
     d->oi_ctrl_down = UIOBJI_INVALID;
     d->oi_ctrl_d2 = UIOBJI_INVALID;
     d->oi_ctrl_dr = UIOBJI_INVALID;
-    UIOBJI_SET_TBL_INVALID(d->oi_tag_set);
-    UIOBJI_SET_TBL_INVALID(d->oi_tag_get);
+    d->oi_pgup = UIOBJI_INVALID;
+    d->oi_pgdown = UIOBJI_INVALID;
 }
 
 void ui_starmap_fill_oi_ctrl(struct starmap_data_s *d)
 {
+    d->oi_scroll = uiobj_add_tb(6, 6, 2, 2, 108, 86, &d->scrollx, &d->scrolly, &d->scrollz, ui_scale);
     d->oi_ctrl_left = uiobj_add_inputkey(MOO_KEY_LEFT | MOO_MOD_CTRL);
     if (MOO_KEY_LEFT != MOO_KEY_KP4) {
         d->oi_ctrl_l2 = uiobj_add_inputkey(MOO_KEY_KP4 | MOO_MOD_CTRL);
@@ -865,10 +1069,8 @@ void ui_starmap_fill_oi_ctrl(struct starmap_data_s *d)
         d->oi_ctrl_d2 = uiobj_add_inputkey(MOO_KEY_KP2 | MOO_MOD_CTRL);
     }
     d->oi_ctrl_dr = uiobj_add_inputkey(MOO_KEY_KP3 | MOO_MOD_CTRL);
-    for (int i = 0; i < PLANET_TAG_NUM; ++i) {
-        d->oi_tag_set[i] = uiobj_add_inputkey((MOO_KEY_1 + i) | MOO_MOD_CTRL);
-        d->oi_tag_get[i] = uiobj_add_inputkey((MOO_KEY_1 + i));
-    }
+    d->oi_pgup = uiobj_add_inputkey(MOO_KEY_PAGEUP);
+    d->oi_pgdown = uiobj_add_inputkey(MOO_KEY_PAGEDOWN);
 }
 
 void ui_starmap_sn0_setup(struct shipnon0_s *sn0, int sd_num, const shipcount_t *ships)
@@ -911,8 +1113,8 @@ void ui_starmap_draw_planetinfo_2(const struct game_s *g, int p1, int p2, int pl
     player_id_t api = (p1 < PLAYER_NUM) ? p1 : p2;
     bool explored = true;
     if (0
-      || ((p1 < PLAYER_NUM) && IS_HUMAN(g, p1) && BOOLVEC_IS0(p->explored, p1))
-      || ((p2 < PLAYER_NUM) && IS_HUMAN(g, p2) && BOOLVEC_IS0(p->explored, p2))
+      || (IS_HUMAN(g, p1) && BOOLVEC_IS0(p->explored, p1))
+      || (IS_HUMAN(g, p2) && BOOLVEC_IS0(p->explored, p2))
     ) {
         explored = false;
     }
@@ -968,27 +1170,41 @@ int ui_starmap_enemy_incoming(const struct game_s *g, player_id_t pi, int i, boo
     return i;
 }
 
-void ui_starmap_scroll(const struct game_s *g, int scrollx, int scrolly, uint8_t scrollz)
+void ui_starmap_common_init(struct game_s *g, struct starmap_data_s *d, player_id_t active_player)
 {
-    int x, y;
-    if (scrollx >= 0) {
-        x = ui_data.starmap.x + scrollx - 54;
-        y = ui_data.starmap.y + scrolly - 43;
-        SETRANGE(x, 0, g->galaxy_maxx - ((108 * ui_scale) / starmap_scale));
-        SETRANGE(y, 0, g->galaxy_maxy - ((86 * ui_scale) / starmap_scale));
-        ui_data.starmap.x2 = x;
-        ui_data.starmap.y2 = y;
-    } else {
-        starmap_scale = scrollz;
-        ui_starmap_set_pos_focus(g, g->active_player);
-    }
+    d->set_pos_focus = ui_starmap_set_pos_focus;
+    d->g = g;
+    d->api = active_player;
+    d->controllable = false;
+    d->show_planet_focus = true;
+    d->anim_delay = 0;
+    d->dist_i = PLANET_NONE;
+    d->scrollx = 0;
+    d->scrolly = 0;
+    d->scrollz = starmap_scale;
 }
 
-/* compute maximum starmap_scale such that map is as big as possible and not clipped */
-void ui_starmap_compute_scale(const struct game_s *g)
+void ui_starmap_common_update_mouse_hover(struct starmap_data_s *d, int16_t oi)
 {
-    int x = ((222 / 2 * ui_scale) / g->galaxy_maxx);
-    int y = ((178 / 2 * ui_scale) / g->galaxy_maxy);
-    starmap_scale = MIN(x, y);
-    SETRANGE(starmap_scale, 1, ui_scale);
+    d->bottom_highlight = -1;
+    if (oi == d->oi_gameopts) {
+        d->bottom_highlight = 0;
+    } else if (oi == d->oi_design) {
+        d->bottom_highlight = 1;
+    } else if (oi == d->oi_fleet) {
+        d->bottom_highlight = 2;
+    } else if (oi == d->oi_map) {
+        d->bottom_highlight = 3;
+    } else if (oi == d->oi_races) {
+        d->bottom_highlight = 4;
+    } else if (oi == d->oi_planets) {
+        d->bottom_highlight = 5;
+    } else if (oi == d->oi_tech) {
+        d->bottom_highlight = 6;
+    } else if (oi == d->oi_next_turn) {
+        d->bottom_highlight = 7;
+    }
+    if (ui_sm_mouseover_focus && d->controllable) {
+        ui_starmap_select_target(d, ui_starmap_cursor_on_star(d, oi));
+    }
 }
