@@ -13,6 +13,7 @@
 #include "kbd.h"
 #include "lbxgfx.h"
 #include "lbxfont.h"
+#include "lib.h"
 #include "log.h"
 #include "types.h"
 #include "uicursor.h"
@@ -26,22 +27,29 @@
 
 /* -------------------------------------------------------------------------- */
 
+static inline bool ui_starmap_orbit_own_in_frange(struct starmap_data_s *d, uint8_t planet_i)
+{
+    const planet_t *p = &d->g->planet[planet_i];
+    return (p->within_frange[d->api] == 1) || ((p->within_frange[d->api] == 2) && d->oo.sn0.have_reserve_fuel);
+}
+
+/* -------------------------------------------------------------------------- */
+
 static void ui_starmap_orbit_own_draw_cb(void *vptr)
 {
     struct starmap_data_s *d = vptr;
     const struct game_s *g = d->g;
-    const planet_t *pf = &g->planet[d->oo.from];
+    const planet_t *pf = &g->planet[d->from];
     const planet_t *pt = &g->planet[g->planet_focus_i[d->api]];
     char buf[0x80];
     STARMAP_LIM_INIT();
 
-    ui_starmap_draw_starmap(d);
-    ui_starmap_draw_button_text(d, true);
+    ui_starmap_draw_basic(d);
     {
         int x, y;
         x = (pf->x - ui_data.starmap.x) * 2 + 23;
         y = (pf->y - ui_data.starmap.y) * 2 + 5;
-        lbxgfx_draw_frame_offs(x, y, ui_data.gfx.starmap.shipbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+        lbxgfx_draw_frame_offs_delay(x, y, !d->anim_delay, ui_data.gfx.starmap.shipbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
     }
     ui_draw_filled_rect(225, 8, 314, 192, 7, ui_scale);
     lbxgfx_draw_frame(224, 5, ui_data.gfx.starmap.move_shi, UI_SCREEN_W, ui_scale);
@@ -52,35 +60,35 @@ static void ui_starmap_orbit_own_draw_cb(void *vptr)
     lbxfont_select_set_12_4(0, 5, 0, 0);
     lbxfont_print_str_center(268, 11, game_str_sm_fleetdep, UI_SCREEN_W, ui_scale);
 
-    if (g->planet_focus_i[d->api] != d->oo.from) {
+    if (g->planet_focus_i[d->api] != d->from) {
         int dist = game_get_min_dist(g, d->api, g->planet_focus_i[d->api]);
         int x0, y0, x1, y1;
         const uint8_t *ctbl;
         uint8_t *gfx;
         x1 = (pt->x - ui_data.starmap.x) * 2 + 8;
         y1 = (pt->y - ui_data.starmap.y) * 2 + 8;
-        lbxgfx_draw_frame_offs(x1, y1, ui_data.gfx.starmap.planbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
+        lbxgfx_draw_frame_offs_delay(x1, y1, !d->anim_delay, ui_data.gfx.starmap.planbord, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
         x0 = (pf->x - ui_data.starmap.x) * 2 + 26;
         y0 = (pf->y - ui_data.starmap.y) * 2 + 8;
-        ctbl = d->oo.in_frange ? colortbl_line_green : colortbl_line_red;
+        ctbl = ui_starmap_orbit_own_in_frange(d, g->planet_focus_i[d->api]) ? colortbl_line_green : colortbl_line_red;
         ui_draw_line_limit_ctbl(x0 + 3, y0 + 1, x1 + 6, y1 + 6, ctbl, 5, ui_data.starmap.line_anim_phase, starmap_scale);
         gfx = ui_data.gfx.starmap.smalship[g->eto[d->api].banner];
         lbxgfx_set_frame_0(gfx);
         lbxgfx_draw_frame_offs(x0, y0, gfx, STARMAP_LIMITS, UI_SCREEN_W, starmap_scale);
-        if (!d->oo.in_frange) {
+        if (!ui_starmap_orbit_own_in_frange(d, g->planet_focus_i[d->api])) {
             if (d->oo.sn0.num < NUM_SHIPDESIGNS) { /* WASBUG MOO1 compares to 7, resulting in text below last ship */
-                sprintf(buf, "%s %i %s", game_str_sm_destoor, dist, game_str_sm_parsfromcc);
+                lib_sprintf(buf, sizeof(buf), "%s %i %s", game_str_sm_destoor, dist, game_str_sm_parsfromcc);
                 lbxfont_select(2, 0, 0, 0);
                 lbxfont_set_gap_h(2);
                 lbxfont_print_str_split(228, 156, 81, buf, 2, UI_SCREEN_W, UI_SCREEN_H, ui_scale);
             } else {
-                sprintf(buf, "%s (%i)", game_str_sm_destoor2, dist);
+                lib_sprintf(buf, sizeof(buf), "%s (%i)", game_str_sm_destoor2, dist);
                 ui_draw_filled_rect(228, 9, 309, 17, 7, ui_scale);
                 lbxfont_print_str_center(268, 11, buf, UI_SCREEN_W, ui_scale);
             }
         } else {
             if ((pt->owner == d->api) && (pf->owner == d->api) && pt->have_stargate && pf->have_stargate) {
-                strcpy(buf, game_str_sm_stargate);
+                lib_strcpy(buf, game_str_sm_stargate, sizeof(buf));
             } else if (d->oo.shiptypenon0numsel > 0) {
                 const shipdesign_t *sd = &(g->srd[d->api].design[0]);
                 int eta, speed = 20;
@@ -92,8 +100,8 @@ static void ui_starmap_orbit_own_draw_cb(void *vptr)
                     }
                 }
                 ++speed;
-                eta = game_calc_eta(g, speed, pt->x, pt->y, pf->x, pf->y);
-                sprintf(buf, "%s %i %s", game_str_sm_eta, eta, (eta == 1) ? game_str_sm_turn : game_str_sm_turns);
+                eta = game_calc_eta_ship(g, speed, pt->x, pt->y, pf->x, pf->y);
+                lib_sprintf(buf, sizeof(buf), "%s %i %s", game_str_sm_eta, eta, (eta == 1) ? game_str_sm_turn : game_str_sm_turns);
             } else {
                 buf[0] = '\0';
             }
@@ -113,14 +121,13 @@ static void ui_starmap_orbit_own_draw_cb(void *vptr)
     }
     for (int i = 0; i < d->oo.sn0.num; ++i) {
         const shipdesign_t *sd = &(g->srd[d->api].design[0]);
-        struct draw_stars_s ds;
         uint8_t *gfx;
         int st;
         ui_draw_filled_rect(227, 22 + i * 26, 259, 46 + i * 26, 0, ui_scale);
         ui_draw_filled_rect(264, 34 + i * 26, 310, 46 + i * 26, 0, ui_scale);
-        ds.xoff1 = 0;
-        ds.xoff2 = 0;
-        ui_draw_stars(227, 22 + i * 26, 0, 32, &ds, ui_scale);
+        ui_data.starmap.stars_xoff1 = 0;
+        ui_data.starmap.stars_xoff2 = 0;
+        ui_draw_stars(227, 22 + i * 26, 0, 32, ui_scale);
         st = d->oo.sn0.type[i];
         gfx = ui_data.gfx.ships[sd[st].look];
         lbxgfx_set_frame_0(gfx);
@@ -129,11 +136,13 @@ static void ui_starmap_orbit_own_draw_cb(void *vptr)
         {
             int y;
             y = 40 + i * 26;
-            if (ui_extra_enabled) {
-                lbxfont_print_num_right(258, y, d->oo.sn0.ships[i], UI_SCREEN_W, ui_scale);
-                y = 24 + i * 26;
-            }
             lbxfont_print_num_right(258, y, d->oo.ships[st], UI_SCREEN_W, ui_scale);
+            if (ui_extra_enabled) {
+                y = 24 + i * 26;
+                lbxfont_select(0, 0x7, 0, 0);
+                lbxfont_print_num_right(258, y, d->oo.sn0.ships[i], UI_SCREEN_W, ui_scale);
+                lbxfont_select(0, 0xd, 0, 0);
+            }
         }
         lbxfont_select_set_12_1(2, 0, 0, 0);
         lbxfont_print_str_center(287, 25 + i * 26, sd[st].name, UI_SCREEN_W, ui_scale);
@@ -147,7 +156,7 @@ static void ui_starmap_orbit_own_draw_cb(void *vptr)
 void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
 {
     bool flag_done = false;
-    int16_t oi_scroll, oi_cancel, oi_accept, oi_search, oi_cycle,
+    int16_t oi_cancel, oi_accept, oi_search,
             oi_f2, oi_f3, oi_f4, oi_f5, oi_f6, oi_f7, oi_f8, oi_f9, oi_f10,
             oi_tbl_p[NUM_SHIPDESIGNS],
             oi_tbl_m[NUM_SHIPDESIGNS],
@@ -155,19 +164,17 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             oi_tbl_n[NUM_SHIPDESIGNS],
             oi_tbl_s[NUM_SHIPDESIGNS]
             ;
-    int16_t scrollx = 0, scrolly = 0, scrollship = 0;
-    uint8_t scrollz = starmap_scale;
+    int16_t scrollship = 0;
     struct starmap_data_s d;
     const fleet_orbit_t *r;
     const shipcount_t *os;
     const uint8_t shiptypes[NUM_SHIPDESIGNS] = { 0, 1, 2, 3, 4, 5 };
 
-    d.g = g;
-    d.api = active_player;
-    d.anim_delay = 0;
-    d.oo.from = g->planet_focus_i[active_player];
+    ui_starmap_common_init(g, &d, active_player);
+    d.from = g->planet_focus_i[active_player];
+    d.controllable = true;
 
-    r = &(g->eto[active_player].orbit[d.oo.from]);
+    r = &(g->eto[active_player].orbit[d.from]);
 
     if (BOOLVEC_IS_CLEAR(r->visible, PLAYER_NUM)) {
         ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
@@ -178,7 +185,6 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
     for (int i = 0; i < NUM_SHIPDESIGNS; ++i) {
         d.oo.ships[i] = os[i];
     }
-    d.oo.in_frange = false;
     ui_starmap_sn0_setup(&d.oo.sn0, NUM_SHIPDESIGNS, d.oo.ships);
 
     uiobj_table_clear();
@@ -189,7 +195,6 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
         STARMAP_UIOBJ_CLEAR_FX(); \
         oi_accept = UIOBJI_INVALID; \
         oi_cancel = UIOBJI_INVALID; \
-        oi_cycle = UIOBJI_INVALID; \
         UIOBJI_SET_TBL5_INVALID(oi_tbl_p, oi_tbl_m, oi_tbl_a, oi_tbl_n, oi_tbl_s); \
     } while (0)
 
@@ -199,73 +204,43 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
     uiobj_set_callback_and_delay(ui_starmap_orbit_own_draw_cb, &d, STARMAP_DELAY);
 
     while (!flag_done) {
-        planet_t *p;
         int16_t oi1, oi2;
-        int cursor_over;
-        p = &g->planet[g->planet_focus_i[active_player]];
         ui_starmap_update_reserve_fuel(g, &d.oo.sn0, d.oo.ships, active_player);
         oi1 = uiobj_handle_input_cond();
         oi2 = uiobj_at_cursor();
         ui_delay_prepare();
         ui_starmap_handle_scrollkeys(&d, oi1);
-        if (oi1 == d.oi_gameopts) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_GAMEOPTS;
+        if (ui_starmap_handle_oi_bottom_buttons(&d, oi1)) {
             flag_done = true;
             ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_design) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_DESIGN;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_fleet) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_FLEET;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_map) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_MAP;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_races) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_RACES;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_planets) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_PLANETS;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_tech) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_TECH;
-            flag_done = true;
-            ui_sound_play_sfx_24();
-        } else if (oi1 == d.oi_next_turn) {
-            ui_data.ui_main_loop_action = UI_MAIN_LOOP_NEXT_TURN;
-            flag_done = true;
+        } else if (ui_starmap_handle_oi_misc(&d, oi1)) {
             ui_sound_play_sfx_24();
         } else if (oi1 == oi_search) {
             ui_sound_play_sfx_24();
             ui_search_set_pos(g, active_player);
         } else if (oi1 == oi_f2) {
             int i;
-            i = d.oo.from;
+            i = d.from;
             do {
                 if (--i < 0) { i = g->galaxy_stars - 1; }
             } while (g->planet[i].owner != active_player);
             g->planet_focus_i[active_player] = i;
             ui_starmap_set_pos_focus(g, active_player);
             ui_sound_play_sfx_24();
-            d.oo.from = i;
+            d.from = i;
             /* flag_have_colony_lbx = false */
             flag_done = true;
             ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
         } else if (oi1 == oi_f3) {
             int i;
-            i = d.oo.from;
+            i = d.from;
             do {
                 i = (i + 1) % g->galaxy_stars;
             } while (g->planet[i].owner != active_player);
             g->planet_focus_i[active_player] = i;
             ui_starmap_set_pos_focus(g, active_player);
             ui_sound_play_sfx_24();
-            d.oo.from = i;
+            d.from = i;
             /* flag_have_colony_lbx = false */
             flag_done = true;
             ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
@@ -276,20 +251,18 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             i = ui_starmap_enemy_incoming(g, active_player, pi, (oi1 == oi_f8));
             if (i != pi) {
                 g->planet_focus_i[active_player] = i;
-                if (!ui_extra_enabled) {
-                    ui_starmap_set_pos_focus(g, active_player);
-                    d.oo.from = i;
-                    /* flag_have_colony_lbx = false */
-                    flag_done = true;
-                    ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
-                }
+                ui_starmap_set_pos_focus(g, active_player);
+                d.from = i;
+                /* flag_have_colony_lbx = false */
+                flag_done = true;
+                ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
             }
         } else if (oi1 == oi_f10) {
             game_save_do_save_i(GAME_SAVE_I_CONTINUE, "Continue", g);
         } else if (oi1 == oi_f4) {
             bool found;
             int i, pi;
-            i = pi = d.oo.from;
+            i = pi = d.from;
             found = false;
             do {
                 i = (i + 1) % g->galaxy_stars;
@@ -303,7 +276,7 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             if (found) {
                 g->planet_focus_i[active_player] = i;
                 ui_starmap_set_pos_focus(g, active_player);
-                d.oo.from = i;
+                d.from = i;
                 ui_sound_play_sfx_24();
                 ui_data.starmap.orbit_player = active_player;
                 flag_done = true;
@@ -311,7 +284,7 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
         } else if (oi1 == oi_f5) {
             bool found;
             int i, pi;
-            i = pi = d.oo.from;
+            i = pi = d.from;
             found = false;
             do {
                 if (--i < 0) { i = g->galaxy_stars - 1; }
@@ -325,17 +298,17 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             if (found) {
                 g->planet_focus_i[active_player] = i;
                 ui_starmap_set_pos_focus(g, active_player);
-                d.oo.from = i;
+                d.from = i;
                 ui_sound_play_sfx_24();
                 ui_data.starmap.orbit_player = active_player;
                 flag_done = true;
             }
         } else if (oi1 == oi_f6) {
             int i;
-            i = ui_starmap_newship_next(g, active_player, d.oo.from);
+            i = ui_starmap_newship_next(g, active_player, d.from);
             g->planet_focus_i[active_player] = i;
             ui_starmap_set_pos_focus(g, active_player);
-            d.oo.from = i;
+            d.from = i;
             ui_sound_play_sfx_24();
             if (BOOLVEC_IS1(g->eto[active_player].orbit[i].visible, active_player)) {
                 ui_data.starmap.orbit_player = active_player;
@@ -345,10 +318,10 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             flag_done = true;
         } else if (oi1 == oi_f7) {
             int i;
-            i = ui_starmap_newship_prev(g, active_player, d.oo.from);
+            i = ui_starmap_newship_prev(g, active_player, d.from);
             g->planet_focus_i[active_player] = i;
             ui_starmap_set_pos_focus(g, active_player);
-            d.oo.from = i;
+            d.from = i;
             ui_sound_play_sfx_24();
             if (BOOLVEC_IS1(g->eto[active_player].orbit[i].visible, active_player)) {
                 ui_data.starmap.orbit_player = active_player;
@@ -379,7 +352,7 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
             for (player_id_t j = PLAYER_0; j < g->players; ++j) {
                 if (oi1 == d.oi_tbl_pl_stars[j][i]) {
                     g->planet_focus_i[active_player] = i;  /* not done on MOO1! */
-                    d.oo.from = i;
+                    d.from = i;
                     ui_data.starmap.orbit_player = j;
                     ui_data.ui_main_loop_action = (j == active_player) ? UI_MAIN_LOOP_ORBIT_OWN_SEL : UI_MAIN_LOOP_ORBIT_EN_SEL;
                     ui_sound_play_sfx_24();
@@ -395,37 +368,22 @@ void ui_starmap_orbit_own(struct game_s *g, player_id_t active_player)
         } else if (oi1 == oi_accept) {
 do_accept:
             ui_sound_play_sfx_24();
-            if ((p->within_frange[active_player] == 1) || ((p->within_frange[active_player] == 2) && d.oo.sn0.have_reserve_fuel)) {
-                game_send_fleet_from_orbit(g, active_player, d.oo.from, g->planet_focus_i[active_player], d.oo.ships, shiptypes, 6);
+            if (ui_starmap_orbit_own_in_frange(&d, g->planet_focus_i[active_player])) {
+                game_send_fleet_from_orbit(g, active_player, d.from, g->planet_focus_i[active_player], d.oo.ships, shiptypes, 6);
                 game_update_visibility(g);
             }
-            if ((!ui_extra_enabled) || BOOLVEC_IS_CLEAR(r->visible, PLAYER_NUM)) {
+            if ((!kbd_is_modifier(MOO_MOD_SHIFT)) || BOOLVEC_IS_CLEAR(r->visible, PLAYER_NUM)) {
                 flag_done = true;
                 ui_data.ui_main_loop_action = UI_MAIN_LOOP_STARMAP;
             } else {
-                bool any_selected;
                 ui_starmap_sn0_setup(&d.oo.sn0, NUM_SHIPDESIGNS, os);
-                any_selected = false;
                 for (int i = 0; i < NUM_SHIPDESIGNS; ++i) {
-                    shipcount_t n;
-                    n = MIN(d.oo.ships[i], os[i]);
-                    d.oo.ships[i] = n;
-                    if (n > 0) {
-                        any_selected = true;
-                    }
+                    SETMIN(d.oo.ships[i], os[i]);
                 }
-                if (!any_selected) {
-                    for (int i = 0; i < NUM_SHIPDESIGNS; ++i) {
-                        d.oo.ships[i] = os[i];
-                    }
-                }
-                g->planet_focus_i[active_player] = d.oo.from;
+                g->planet_focus_i[active_player] = d.from;
             }
-        } else if (oi1 == oi_scroll) {
-            ui_starmap_scroll(g, scrollx, scrolly, scrollz);
         }
         ui_starmap_handle_oi_ctrl(&d, oi1);
-        ui_starmap_handle_tag(&d, oi1, false);
         for (int i = 0; i < g->galaxy_stars; ++i) {
             if (oi1 == d.oi_tbl_stars[i]) {
                 if (ui_extra_enabled && (oi_accept != UIOBJI_INVALID) && (g->planet_focus_i[active_player] == i)) {
@@ -437,28 +395,14 @@ do_accept:
                 break;
             }
         }
-        cursor_over = -1;
-        for (int i = 0; i < d.oo.sn0.num; ++i) {
-            if (0
-              || (oi2 == oi_tbl_p[i])
-              || (oi2 == oi_tbl_m[i])
-              || (oi2 == oi_tbl_a[i])
-              || (oi2 == oi_tbl_n[i])
-            ) {
-                cursor_over = i;
-                break;
-            }
-        }
-        if (oi1 == oi_cycle) {
-            if (++cursor_over >= d.oo.sn0.num) {
-                cursor_over = 0;
-            }
-            uiobj_set_focus(oi_tbl_m[cursor_over]);
-        }
+
         for (int i = 0; i < d.oo.sn0.num; ++i) {
             int si, per10num;
             si = d.oo.sn0.type[i];
             per10num = os[si] / 10;
+            if (kbd_is_modifier(MOO_MOD_CTRL)) {
+                per10num = 1;
+            }
             SETMAX(per10num, 1);
             if ((oi1 == oi_tbl_p[i]) || ((oi1 == oi_tbl_s[i]) && ((scrollship > 0) != ui_mwi_counter))) {
                 shipcount_t t;
@@ -500,33 +444,7 @@ do_accept:
             }
         }
         if (!flag_done) {
-            p = &g->planet[g->planet_focus_i[active_player]];
-            if (1
-              && (g->planet_focus_i[active_player] != d.oo.from)
-              && ((p->within_frange[active_player] == 1) || ((p->within_frange[active_player] == 2) && d.oo.sn0.have_reserve_fuel))
-            ) {
-                d.oo.in_frange = true;
-            } else {
-                d.oo.in_frange = false;
-            }
-            d.bottom_highlight = -1;
-            if (oi2 == d.oi_gameopts) {
-                d.bottom_highlight = 0;
-            } else if (oi2 == d.oi_design) {
-                d.bottom_highlight = 1;
-            } else if (oi2 == d.oi_fleet) {
-                d.bottom_highlight = 2;
-            } else if (oi2 == d.oi_map) {
-                d.bottom_highlight = 3;
-            } else if (oi2 == d.oi_races) {
-                d.bottom_highlight = 4;
-            } else if (oi2 == d.oi_planets) {
-                d.bottom_highlight = 5;
-            } else if (oi2 == d.oi_tech) {
-                d.bottom_highlight = 6;
-            } else if (oi2 == d.oi_next_turn) {
-                d.bottom_highlight = 7;
-            }
+            ui_starmap_common_update_mouse_hover(&d, oi2);
             ui_starmap_orbit_own_draw_cb(&d);
             uiobj_table_clear();
             UIOBJ_CLEAR_LOCAL();
@@ -539,24 +457,31 @@ do_accept:
             oi_f8 = uiobj_add_inputkey(MOO_KEY_F8);
             oi_f9 = uiobj_add_inputkey(MOO_KEY_F9);
             oi_f10 = uiobj_add_inputkey(MOO_KEY_F10);
-            ui_starmap_fill_oi_tbls(&d);
+            if (!ui_sm_explicit_cursor_context || kbd_is_modifier(MOO_MOD_ALT)) {
+                ui_starmap_fill_oi_tbls(&d);
+                ui_cursor_setup_area(2, &ui_cursor_area_tbl[3]);
+            } else {
+                ui_cursor_setup_area(2, &ui_cursor_area_tbl[1]);
+            }
             ui_starmap_fill_oi_tbl_stars(&d);
             oi_cancel = uiobj_add_t0(227, 180, "", ui_data.gfx.starmap.reloc_bu_cancel, MOO_KEY_ESCAPE);
-            if (d.oo.in_frange && d.oo.shiptypenon0numsel) {
+            if (1
+              && ((g->planet_focus_i[active_player] != d.from) || ui_extra_enabled)
+              && ui_starmap_orbit_own_in_frange(&d, g->planet_focus_i[active_player])
+              && d.oo.shiptypenon0numsel) {
                 oi_accept = uiobj_add_t0(271, 180, "", ui_data.gfx.starmap.reloc_bu_accept, MOO_KEY_SPACE);
             }
-            oi_cycle = uiobj_add_inputkey(MOO_KEY_TAB);
-            oi_scroll = uiobj_add_tb(6, 6, 2, 2, 108, 86, &scrollx, &scrolly, &scrollz, ui_scale);
             oi_search = uiobj_add_inputkey(MOO_KEY_SLASH);
             ui_starmap_fill_oi_ctrl(&d);
             for (int i = 0; i < d.oo.sn0.num; ++i) {
-                oi_tbl_p[i] = uiobj_add_t0(288, 35 + i * 26, "", ui_data.gfx.starmap.move_but_p, (cursor_over == i) ? MOO_KEY_GREATER : MOO_KEY_UNKNOWN);
-                oi_tbl_m[i] = uiobj_add_t0(277, 35 + i * 26, "", ui_data.gfx.starmap.move_but_m, (cursor_over == i) ? MOO_KEY_LESS : MOO_KEY_UNKNOWN);
+                oi_tbl_p[i] = uiobj_add_t0(288, 35 + i * 26, "", ui_data.gfx.starmap.move_but_p, MOO_KEY_UNKNOWN);
+                oi_tbl_m[i] = uiobj_add_t0(277, 35 + i * 26, "", ui_data.gfx.starmap.move_but_m, MOO_KEY_UNKNOWN);
                 oi_tbl_a[i] = uiobj_add_t0(299, 35 + i * 26, "", ui_data.gfx.starmap.move_but_a, MOO_KEY_UNKNOWN);
                 oi_tbl_n[i] = uiobj_add_t0(265, 35 + i * 26, "", ui_data.gfx.starmap.move_but_n, MOO_KEY_UNKNOWN);
                 oi_tbl_s[i] = uiobj_add_mousewheel(227, 22 + i * 26, 319, 46 + i * 26, &scrollship);
             }
             ui_starmap_add_oi_bottom_buttons(&d);
+            ui_starmap_add_oi_misc(&d);
             d.oi_tech = UIOBJI_INVALID;
             d.oi_next_turn = UIOBJI_INVALID;
             ui_draw_finish();
@@ -566,5 +491,5 @@ do_accept:
     uiobj_unset_callback();
     uiobj_table_clear();
     uiobj_set_help_id(-1);
-    g->planet_focus_i[active_player] = d.oo.from;
+    g->planet_focus_i[active_player] = d.from;
 }
